# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import attr

from contextlib import contextmanager
from io import BytesIO
import grpc
import os
from pathlib import Path
import shutil
import tarfile

import pytest
from mercurial_testhelpers import (
    as_bytes,
)


from hgext3rd.heptapod.branch import (
    get_default_gitlab_branch,
    read_gitlab_typed_refs,
    set_default_gitlab_branch,
)
from hgext3rd.heptapod.keep_around import (
    create_keep_around,
    iter_keep_arounds,
)

from mercurial_testhelpers import RepoWrapper

from hgitaly.license_detector import license_content
from hgitaly.revision import ZERO_SHA
from hgitaly.stream import WRITE_BUFFER_SIZE
from hgitaly.testing import TEST_DATA_DIR
from hgitaly.testing.bundle import list_bundle_contents
from hgitaly.workdir import (
    ROSTER_FILE_NAME as WDS_ROSTER_FILE_NAME,
    working_directory,
)

from hgitaly.stub.shared_pb2 import Repository
from hgitaly import stream

from hgitaly.stub.repository_pb2 import (
    BackupCustomHooksRequest,
    CreateBundleRequest,
    CreateRepositoryFromBundleRequest,
    CreateRepositoryRequest,
    CreateBundleFromRefListRequest,
    FetchBundleRequest,
    FindLicenseRequest,
    FindLicenseResponse,
    FindMergeBaseRequest,
    GetArchiveRequest,
    GetCustomHooksRequest,
    HasLocalBranchesRequest,
    ObjectFormatRequest,
    RemoveRepositoryRequest,
    RepositoryExistsRequest,
    RepositorySizeRequest,
    RestoreCustomHooksRequest,
    SearchFilesByNameRequest,
    SearchFilesByContentRequest,
    SearchFilesByContentResponse,
    SetCustomHooksRequest,
    WriteRefRequest,
)
from hgitaly.stub.repository_pb2_grpc import RepositoryServiceStub
from hgitaly.stub.shared_pb2 import (
    ObjectFormat,
)
from mercurial import (
    error,
    node as node_mod,
    phases,
)
from .fixture import ServiceFixture

parametrize = pytest.mark.parametrize

ALREADY_EXISTS = grpc.StatusCode.ALREADY_EXISTS
INTERNAL = grpc.StatusCode.INTERNAL
INVALID_ARGUMENT = grpc.StatusCode.INVALID_ARGUMENT
NOT_FOUND = grpc.StatusCode.NOT_FOUND

END_OF_MATCH = SearchFilesByContentResponse(end_of_match=True)


@attr.s
class RepositoryFixture(ServiceFixture):

    stub_cls = RepositoryServiceStub
    auto_cleanup = attr.ib(default=True)
    """cannot be overridden by class attribute"""

    def exists(self):
        return self.stub.RepositoryExists(
            RepositoryExistsRequest(repository=self.grpc_repo)).exists

    def size(self, **kw):
        grpc_repo = kw.pop('grpc_repo', self.grpc_repo)
        return self.stub.RepositorySize(
            RepositorySizeRequest(repository=grpc_repo, **kw)).size

    def has_local_branches(self):
        return self.stub.HasLocalBranches(
            HasLocalBranchesRequest(repository=self.grpc_repo)).value

    def write_ref(self, ref, target, **kw):
        self.stub.WriteRef(WriteRefRequest(
            repository=self.grpc_repo,
            ref=ref,
            revision=target,
            **kw
        ))

    @contextmanager
    def get_archive_tarfile(self, commit_id, path=b''):
        with BytesIO() as fobj:
            for chunk_index, chunk_response in enumerate(
                    self.stub.GetArchive(GetArchiveRequest(
                        repository=self.grpc_repo,
                        format=GetArchiveRequest.Format.Value('TAR'),
                        commit_id=commit_id,
                        path=path,
                        prefix='archive-dir',
                    ))):
                fobj.write(chunk_response.data)

            fobj.seek(0)
            with tarfile.open(fileobj=fobj) as tarf:
                yield tarf, chunk_index + 1

    def find_merge_base(self, revisions):
        return self.stub.FindMergeBase(FindMergeBaseRequest(
            repository=self.grpc_repo,
            revisions=revisions,
        )).base.encode('ascii')  # needed for comparison with `changectx.hex()`

    def create_repository(self, rel_path, storage_name='default',
                          **req_kwargs):
        self.grpc_repo = Repository(relative_path=rel_path,
                                    storage_name=storage_name)
        self.stub.CreateRepository(CreateRepositoryRequest(
            repository=self.grpc_repo, **req_kwargs))
        self.repo_wrapper = self.make_repo_wrapper(rel_path,
                                                   storage_name=storage_name)

    def search_files_by_name(self, ref=b'branch/default', **kwargs):
        return [path
                for resp in self.stub.SearchFilesByName(
                        SearchFilesByNameRequest(
                            repository=self.grpc_repo,
                            ref=ref,
                            **kwargs,
                        ))
                for path in resp.files]

    def search_files_by_content(self, ref=b'branch/default',
                                aggregate_splitlines=True,
                                **kwargs):
        stream = self.stub.SearchFilesByContent(SearchFilesByContentRequest(
            repository=self.grpc_repo,
            ref=ref,
            **kwargs,
        ))
        if not aggregate_splitlines:
            return list(stream)

        res = []
        match_data = BytesIO()
        for resp in stream:
            if resp.end_of_match:
                res.append(match_data.getvalue().split(b'\n'))
                match_data = BytesIO()  # deliberately not as in tested code
                res.append(resp)
            else:
                match_data.write(resp.match_data)
        return res

    def restore_custom_hooks(self, tarball_path,
                             grpc_repo=None,
                             nb_chunks=2,
                             **kw):
        if grpc_repo is None:
            grpc_repo = self.grpc_repo

        return self.stub.RestoreCustomHooks(
            RestoreCustomHooksRequest(repository=grpc_repo, data=chunk)
            for chunk in stream.slice_binary_file(tarball_path, nb_chunks)
        )

    def set_custom_hooks(self, tarball_path,
                         grpc_repo=None,
                         nb_chunks=2,
                         **kw):
        if grpc_repo is None:
            grpc_repo = self.grpc_repo

        return self.stub.SetCustomHooks(
            SetCustomHooksRequest(repository=grpc_repo, data=chunk)
            for chunk in stream.slice_binary_file(tarball_path, nb_chunks)
        )

    def backup_custom_hooks(self, save_path):
        with open(save_path, 'wb') as fobj:
            for chunk in self.stub.BackupCustomHooks(BackupCustomHooksRequest(
                    repository=self.grpc_repo)):
                fobj.write(chunk.data)

    def get_custom_hooks(self, save_path):
        with open(save_path, 'wb') as fobj:
            for chunk in self.stub.GetCustomHooks(GetCustomHooksRequest(
                    repository=self.grpc_repo)):
                fobj.write(chunk.data)

    def remove_repository(self, grpc_repo=None):
        if grpc_repo is None:
            grpc_repo = self.grpc_repo

        return self.stub.RemoveRepository(RemoveRepositoryRequest(
            repository=grpc_repo))

    def object_format(self, **kw):
        grpc_repo = kw.pop('grpc_repo', self.grpc_repo)
        return self.stub.ObjectFormat(ObjectFormatRequest(repository=grpc_repo,
                                                          **kw))

    def create_bundle(self, save_path):
        with open(save_path, 'wb') as fobj:
            for chunk in self.stub.CreateBundle(CreateBundleRequest(
                    repository=self.grpc_repo)):
                fobj.write(chunk.data)

    def fetch_bundle(self, bundle_path, grpc_repo, nb_chunks=2, **kw):
        return self.stub.FetchBundle(
            FetchBundleRequest(repository=grpc_repo, data=chunk)
            for chunk in stream.slice_binary_file(bundle_path, nb_chunks)
        )

    def create_repository_from_bundle(self, bundle_path, grpc_repo,
                                      nb_chunks=2):
        """Call CreateRepositoryFromBundle.

        Makes several chunks by default to test that aggregation works.
        """
        return self.stub.CreateRepositoryFromBundle(
            CreateRepositoryFromBundleRequest(repository=grpc_repo, data=chunk)
            for chunk in stream.slice_binary_file(bundle_path, nb_chunks)
        )

    def find_license(self):
        return self.stub.FindLicense(FindLicenseRequest(
            repository=self.grpc_repo))

    def create_bundle_from_ref_list(self, bundle_path, ref_patterns,
                                    nb_chunks=2,
                                    without_repository=False):
        """Call CreateBundleFromRefList.

        Makes several chunks by default to test that aggregation works.
        """
        def get_request_iter(patterns):
            first_req = True
            for chunk in stream.split_batches(patterns, nb_chunks):
                req_args = dict(patterns=chunk)
                if first_req and not without_repository:
                    first_req = False
                    req_args['repository'] = self.grpc_repo
                yield CreateBundleFromRefListRequest(**req_args)

        with open(bundle_path, 'wb') as fobj:
            for chunk in self.stub.CreateBundleFromRefList(
                    get_request_iter(ref_patterns)):
                fobj.write(chunk.data)

    def assert_compare_repos(self, target_repo_path):
        """Compare the fixture repo with another one.
        """
        wrapper = self.repo_wrapper
        orig = wrapper.repo.unfiltered()
        target = self.make_repo_wrapper(target_repo_path).repo.unfiltered()
        for rev_str in ('draft()', 'public()', 'obsolete()'):
            assert (
                [r for r in orig.revs(rev_str)]
                ==
                [r for r in target.revs(rev_str)])

        # assert empty incoming/outgoing changes between orig and target repo
        no_changes = 1  # code returned when no changes to transmit
        # XXX: extra args bundle, force should not be required once
        # upstream Mercurial use opts.get() instead of [] access
        incoming = wrapper.command('incoming', target.root,
                                   bundle=None, force=None)
        outgoing = wrapper.command('outgoing', target.root,
                                   bundle=None, force=None)
        assert incoming == no_changes
        assert outgoing == no_changes

        # assert branches
        orig_br = [br[0] for br in orig.branchmap().iterbranches()]
        target_br = [br[0] for br in target.branchmap().iterbranches()]
        assert orig_br == target_br

        # assert tags
        assert orig.tags() == target.tags()


@pytest.fixture
def fixture_with_repo(grpc_channel, server_repos_root):
    with RepositoryFixture(grpc_channel, server_repos_root) as fixture:
        yield fixture


@pytest.fixture
def fixture_without_repo(grpc_channel, server_repos_root):
    with RepositoryFixture(grpc_channel, server_repos_root,
                           with_repo=False) as fixture:
        yield fixture


def test_repository_exists(fixture_with_repo):
    fixture = fixture_with_repo
    assert fixture.exists()

    # directory exists but is not a Mercurial repo
    shutil.rmtree(fixture.repo_wrapper.path / '.hg')
    assert not fixture.exists()

    # directory does not exist
    fixture.grpc_repo.relative_path = 'does/not/exist'
    assert not fixture.exists()

    # storage does not exist
    fixture.grpc_repo.storage_name = 'dream'
    with pytest.raises(grpc.RpcError) as exc_info:
        fixture.exists()
    assert exc_info.value.code() == INVALID_ARGUMENT

    # missing repo argument
    fixture.grpc_repo = None
    with pytest.raises(grpc.RpcError) as exc_info:
        fixture.exists()
    assert exc_info.value.code() == INVALID_ARGUMENT


def test_repository_size(fixture_with_repo):
    fixture = fixture_with_repo
    size = fixture.size()
    # current value is 16 KiB, but let's not depend on it exactly
    assert size < 200
    assert size > 0

    # directory exists but is not a Mercurial repo
    shutil.rmtree(fixture.repo_wrapper.path / '.hg')
    with pytest.raises(grpc.RpcError) as exc_info:
        fixture.size()
    assert exc_info.value.code() == NOT_FOUND

    # directory does not exist
    with pytest.raises(grpc.RpcError) as exc_info:
        fixture.size()
    assert exc_info.value.code() == NOT_FOUND

    # storage does not exist
    fixture.grpc_repo.storage_name = 'dream'
    with pytest.raises(grpc.RpcError) as exc_info:
        fixture.size()
    assert exc_info.value.code() == INVALID_ARGUMENT

    # missing repo argument
    with pytest.raises(grpc.RpcError) as exc_info:
        fixture.size(grpc_repo=None)
    assert exc_info.value.code() == INVALID_ARGUMENT


def test_has_local_branches(fixture_with_repo):
    fixture = fixture_with_repo
    wrapper = fixture.repo_wrapper

    assert not fixture.has_local_branches()
    wrapper.write_commit('foo')
    assert fixture.has_local_branches()

    wrapper.command('commit', message=b"closing the only head!",
                    close_branch=True)

    assert not fixture.has_local_branches()

    # missing repo argument
    fixture.grpc_repo = None
    with pytest.raises(grpc.RpcError) as exc_info:
        fixture.has_local_branches()
    assert exc_info.value.code() == INVALID_ARGUMENT


def test_write_ref(fixture_with_repo):
    fixture = fixture_with_repo
    wrapper = fixture.repo_wrapper
    repo = wrapper.repo

    for ref, revision in ((b'refs/heads/something', b'dead01234cafe'),
                          (b'HEAD', b'topic/default/wont-last'),
                          (b'refs/keep-around/not-a-sha', b'not-a-sha'),
                          (b'refs/keep-around/feca01eade', b'cafe01dead'),
                          ):
        with pytest.raises(grpc.RpcError) as exc_info:
            fixture.write_ref(ref, revision)
        assert exc_info.value.code() == INVALID_ARGUMENT

    sha = wrapper.commit_file('foo').hex()
    sha1 = wrapper.commit_file('foo', branch='other').hex()

    mr_ref = b'refs/merge-requests/12/head'
    to_write = {b'refs/merge-requests/12/head': sha,
                b'refs/pipelines/98192': b'refs/heads/branch/default',
                b'refs/environments/13/deployments/9826': b'branch/default',
                }
    for ref_path, target in to_write.items():
        fixture.write_ref(ref_path, target)
    assert read_gitlab_typed_refs(repo, 'special-refs') == {
        ref_path[5:]: sha for ref_path in to_write.keys()
    }

    # old_revision logic
    with pytest.raises(grpc.RpcError) as exc_info:
        # ref already exists
        fixture.write_ref(mr_ref, sha1, old_revision=ZERO_SHA)
    assert exc_info.value.code() == INTERNAL

    with pytest.raises(grpc.RpcError) as exc_info:
        # ref not pointing where specified
        fixture.write_ref(mr_ref, sha1, old_revision=b"some-other-value")
    assert exc_info.value.code() == INTERNAL

    with pytest.raises(grpc.RpcError) as exc_info:
        # ref not pointing where specified
        fixture.write_ref(mr_ref, sha1, old_revision=sha1)
    assert exc_info.value.code() == INTERNAL

    fixture.write_ref(mr_ref, sha1, old_revision=sha)
    assert read_gitlab_typed_refs(repo, 'special-refs')[mr_ref[5:]] == sha1

    # Keep-arounds. let's have a pre-existing one for good measure
    existing_ka = b'c8c3ae298f5549a0eb0c28225dcc4f6937b959a8'
    create_keep_around(repo, existing_ka)
    fixture.write_ref(b'refs/keep-around/' + sha, sha)
    assert set(iter_keep_arounds(repo)) == {sha, existing_ka}


def test_write_special_refs_exceptions(fixture_with_repo):
    fixture = fixture_with_repo
    wrapper = fixture.repo_wrapper
    wrapper.commit_file('foo')

    (wrapper.path / '.hg' / 'store' / 'gitlab.special-refs').write_text(
        'invalid')
    # TODO handle exception once we've started raising them. See big
    # `try ... except` block in `repository_service`.
    fixture.write_ref(b'refs/merge-requests/12/head', b'76ac23fe' * 5)


def test_get_archive(fixture_with_repo, tmpdir):
    # repo_stub = RepositoryServiceStub(grpc_channel)
    # wrapper, grpc_repo = make_empty_repo(server_repos_root)
    fixture = fixture_with_repo
    wrapper = fixture.repo_wrapper

    ctx = wrapper.write_commit('foo', content="Foo")
    (wrapper.path / 'sub').mkdir()
    ctx2 = wrapper.write_commit('sub/bar', content="Bar")

    node_str = ctx.hex().decode()
    with fixture.get_archive_tarfile(node_str) as (tarf, _nb):
        assert set(tarf.getnames()) == {'archive-dir/.hg_archival.txt',
                                        'archive-dir/foo'}

        extract_dir = tmpdir.join('extract')
        tarf.extractall(path=extract_dir)

        metadata_lines = extract_dir.join('archive-dir',
                                          '.hg_archival.txt').readlines()

        assert 'node: %s\n' % node_str in metadata_lines
        assert extract_dir.join('archive-dir', 'foo').read() == "Foo"

    node2_str = ctx2.hex().decode()
    with fixture.get_archive_tarfile(node2_str) as (tarf, _nb):
        assert set(tarf.getnames()) == {'archive-dir/.hg_archival.txt',
                                        'archive-dir/foo',
                                        'archive-dir/sub/bar'}

        extract_dir = tmpdir.join('extract-2')
        tarf.extractall(path=extract_dir)

        metadata_lines = extract_dir.join('archive-dir',
                                          '.hg_archival.txt').readlines()

        assert 'node: %s\n' % node2_str in metadata_lines
        assert extract_dir.join('archive-dir', 'sub', 'bar').read() == "Bar"

    with fixture.get_archive_tarfile(node2_str, path=b'sub') as (tarf, _nb):
        assert tarf.getnames() == ['archive-dir/sub/bar']

        extract_dir = tmpdir.join('extract-sub')
        tarf.extractall(path=extract_dir)
        assert extract_dir.join('archive-dir', 'sub', 'bar').read() == "Bar"

    with pytest.raises(grpc.RpcError) as exc_info:
        # entering the context is needed to actually perform the RPC call
        fixture.get_archive_tarfile(node2_str, path=b'/etc/passwd').__enter__()
    assert exc_info.value.code() == INVALID_ARGUMENT


def test_get_archive_multiple_chunks(fixture_with_repo, tmpdir):
    fixture = fixture_with_repo
    wrapper = fixture.repo_wrapper

    # we can't just override the environment variable: it's read at module
    # import time.
    large_content = "Foo" * 200000
    ctx = wrapper.write_commit('foo', content=large_content)
    node_str = ctx.hex().decode()
    with fixture.get_archive_tarfile(node_str) as (tarf, count):
        assert count > 1
        assert set(tarf.getnames()) == {'archive-dir/.hg_archival.txt',
                                        'archive-dir/foo'}

        extract_dir = tmpdir.join('extract')
        tarf.extractall(path=extract_dir)

        metadata_lines = extract_dir.join('archive-dir',
                                          '.hg_archival.txt').readlines()

        assert 'node: %s\n' % node_str in metadata_lines
        assert extract_dir.join('archive-dir', 'foo').read() == large_content

    del large_content


def test_find_merge_base(fixture_with_repo, tmpdir):
    fixture = fixture_with_repo
    wrapper = fixture.repo_wrapper
    # repo structure:
    #
    #   o 2 add animal (branch/stable)
    #   |
    #   | 1 add bar
    #   |/
    #   |
    #   o 0 add foo    o 3 tut
    #
    ctx0 = wrapper.write_commit('foo')
    sha0 = ctx0.hex()
    sha1 = wrapper.write_commit('bar').hex()
    sha2 = wrapper.write_commit('animal', branch='stable', parent=ctx0).hex()
    # commting new root, to test no gca case
    sha3 = wrapper.commit_file('tut', branch='other',
                               parent=node_mod.nullid).hex()

    # Actual test
    assert fixture.find_merge_base([sha1, sha2]) == sha0
    assert fixture.find_merge_base([b'branch/stable', sha1]) == sha0

    # when no merge base (gca) found
    assert fixture.find_merge_base([sha0, sha3]) == b''

    # test with invalid_argument, as it requires minimum 2 revisions
    with pytest.raises(grpc.RpcError) as exc_info:
        fixture.find_merge_base([sha0])
    assert exc_info.value.code() == INVALID_ARGUMENT
    assert (
        exc_info.value.details()
        ==
        'at least 2 revisions are required'
    )

    sha_not_exists = b'deadnode' * 5
    assert fixture.find_merge_base([sha0, sha_not_exists]) == b''

    # cases with obsolescence
    wrapper.update(sha2)
    wrapper.amend_file('animal', message='amended animal').hex()

    assert fixture.find_merge_base([sha1, sha2]) == sha0

    wrapper.update(sha0)
    wrapper.amend_file('foo', message='amended foo')
    assert fixture.find_merge_base([sha1, sha2]) == sha0

    # cases with more than 2 changesets
    # (in a previous version, only the 2 first arguments would have been
    # considered, giving wrong results but no error)
    assert fixture.find_merge_base([sha1, sha1, sha2]) == sha0
    assert fixture.find_merge_base([sha1, sha2, sha3]) == b''


def test_create_repository(fixture_without_repo):
    fixture = fixture_without_repo
    rel_path = 'sample_repo'

    # try to instantiate wrapper before repo creation
    with pytest.raises(error.RepoError) as exc_info:
        fixture.make_repo_wrapper(rel_path)
    assert b''.join(exc_info.value.args).endswith(b'sample_repo not found')

    fixture.create_repository(rel_path)
    # instantiating wrapper to check successful repo creation
    assert fixture.repo_wrapper.repo.path.endswith(b'sample_repo/.hg')

    # As of Gitaly 14.6, attempt to create existing repo is an error.
    with pytest.raises(grpc.RpcError) as exc_info:
        fixture.create_repository(rel_path)
    assert exc_info.value.code() == ALREADY_EXISTS
    assert 'exists already' in exc_info.value.details()

    # when storage doesn't exist
    with pytest.raises(grpc.RpcError) as exc_info:
        fixture.create_repository(rel_path, storage_name='cargoship')
    assert exc_info.value.code() == INVALID_ARGUMENT
    assert 'no such storage' in exc_info.value.details()

    # test with a broken symlink (which points to itself)
    # This used to be a way to test failure in `hg init`, now should be
    # refused as already existing path (see Gitaly Comparison tests)
    brepo_name = "myrepo_a_broken_symlink"
    path = fixture.repo_path(brepo_name)
    path.symlink_to(path)
    with pytest.raises(grpc.RpcError) as exc_info:
        fixture.create_repository(brepo_name)
    assert exc_info.value.code() == ALREADY_EXISTS
    assert 'exists already' in exc_info.value.details()

    # using a broken symlink for the whole storage directory
    # to pass the check for existing repo and still fail in `hg init`
    rel_path = 'creation_error'
    storage_path = fixture.storage_path()
    try:
        shutil.rmtree(storage_path)
        storage_path.symlink_to(storage_path)
        with pytest.raises(grpc.RpcError) as exc_info:
            fixture.create_repository(rel_path)
        assert exc_info.value.code() == INTERNAL
    finally:
        storage_path.unlink()
        storage_path.mkdir(exist_ok=True)


def test_create_repository_default_branch(fixture_without_repo):
    fixture = fixture_without_repo
    rel_path = 'repo_with_default'
    fixture.create_repository(rel_path, default_branch=b'branch/main')
    wrapper = fixture.make_repo_wrapper(rel_path)
    assert get_default_gitlab_branch(wrapper.repo) == b'branch/main'
    fixture.additional_repo(rel_path)  # register cleanup


def test_search_files_by_name(fixture_with_repo):
    fixture = fixture_with_repo
    wrapper = fixture.repo_wrapper

    ctx0 = wrapper.write_commit('afoo', message="Some foo")
    sub = (wrapper.path / 'sub')
    sub.mkdir()
    (sub / 'bar').write_text('bar content')
    (sub / 'ba2').write_text('ba2 content')
    # TODO OS indep for paths (actually TODO make wrapper.commit easier to
    # use, e.g., check how to make it accept patterns)
    wrapper.commit(rel_paths=['sub/bar', 'sub/ba2'],
                   message="zebar", add_remove=True)

    wrapper.write_commit('animals', message="A branch without subdir",
                         branch='other', parent=ctx0)

    search = fixture.search_files_by_name
    assert search(filter='/bar', query='.') == [b'sub/bar']
    assert search(filter='^sub/', query='.') == [b'sub/ba2', b'sub/bar']
    # matching in subdirectory only
    assert search(filter='a', query='sub') == [b'sub/ba2', b'sub/bar']
    assert search(filter='b?r$', query='sub') == [b'sub/bar']

    # without filter
    assert search(query='.') == [b'afoo', b'sub/ba2', b'sub/bar']
    assert search(query='sub') == [b'sub/ba2', b'sub/bar']

    # offset and limit
    assert search(query='.', limit=2) == [b'afoo', b'sub/ba2']
    assert search(query='.', limit=1, offset=1) == [b'sub/ba2']

    # from a different ref
    assert search(filter='bar', query='.', ref=b'branch/other') == []
    assert search(filter='.*', query='.', ref=b'branch/other') == [b'afoo',
                                                                   b'animals']

    # query is mandatory
    with pytest.raises(grpc.RpcError) as exc_info:
        search(filter='/ba')
    assert exc_info.value.code() == INVALID_ARGUMENT

    # problems with regexp
    with pytest.raises(grpc.RpcError) as exc_info:
        search(filter='\\', query='.')
    assert exc_info.value.code() == INVALID_ARGUMENT

    with pytest.raises(grpc.RpcError) as exc_info:
        search(filter='a' * 1001, query='.')
    assert exc_info.value.code() == INVALID_ARGUMENT

    # unknown ref gives back an empty list of files, not an error
    assert search(ref=b'unknown', query='.') == []


def test_search_files_by_content(fixture_with_repo):
    fixture = fixture_with_repo
    wrapper = fixture.repo_wrapper

    wrapper.write_commit('afoo', message="Some foo")
    sub = (wrapper.path / 'sub')
    sub.mkdir()
    (sub / 'bar').write_text('line1\nline2\nbar content\nline4\nline5')
    (sub / 'ba2').write_text('line1\nba2 content')
    (sub / 'ba3').write_text('ba3 content\nline2')
    # this one has Windows and MacOS Classic line endings:
    (sub / 'ba4').write_text('l1\nl2\r\nl3\rba4 content\nl5')
    (sub / 'ba5').write_text('m1\nm2\nm3\nm4\nm5\nm6\nba5 content\n')
    long_content = 'very large content\n' * 3000
    shutil.copy2(TEST_DATA_DIR / 'backup_additional_no_git.tar',
                 wrapper.path / 'bin')
    (wrapper.path / 'large').write_text(long_content)

    wrapper.commit(rel_paths=['bin', 'sub/bar', 'sub/ba2', 'sub/ba3',
                              'sub/ba4', 'sub/ba5', 'large'],
                   message="zebar", add_remove=True)
    search = fixture.search_files_by_content

    # cases not matching
    assert not search(query='no match for this one')
    assert not search(query='PaxHeader')  # tarfile is excluded
    assert not search(ref=b"no/such/fref", query='foo')

    # simple case: match, with two lines of content before and after
    assert search(query="^bar.c") == [
        [b'branch/default:sub/bar\x001\x00line1',
         b'branch/default:sub/bar\x002\x00line2',
         b'branch/default:sub/bar\x003\x00bar content',
         b'branch/default:sub/bar\x004\x00line4',
         b'branch/default:sub/bar\x005\x00line5',
         b'',
         ],
        END_OF_MATCH,
    ]

    # with only one line if file before and after match
    assert search(query="^ba2.c") == [
        [b'branch/default:sub/ba2\x001\x00line1',
         b'branch/default:sub/ba2\x002\x00ba2 content',
         b'',
         ],
        END_OF_MATCH,
    ]
    assert search(query="^ba3.c") == [
        [b'branch/default:sub/ba3\x001\x00ba3 content',
         b'branch/default:sub/ba3\x002\x00line2',
         b'',
         ],
        END_OF_MATCH,
    ]

    # more than two lines before match
    assert search(query="^ba4.c") == [
        [b'branch/default:sub/ba4\x002\x00l2',
         b'branch/default:sub/ba4\x003\x00l3',
         b'branch/default:sub/ba4\x004\x00ba4 content',
         b'branch/default:sub/ba4\x005\x00l5',
         b'',
         ],
        END_OF_MATCH,
    ]

    # two matches, with overlapping context or not
    assert search(query="^l1|ba4.c") == [
        [b'branch/default:sub/ba4\x001\x00l1',
         b'branch/default:sub/ba4\x002\x00l2',
         b'branch/default:sub/ba4\x003\x00l3',
         b'branch/default:sub/ba4\x004\x00ba4 content',
         b'branch/default:sub/ba4\x005\x00l5',
         b'',
         ],
        END_OF_MATCH,
    ]
    assert search(query="^m1|ba5.c") == [
        [b'branch/default:sub/ba5\x001\x00m1',
         b'branch/default:sub/ba5\x002\x00m2',
         b'branch/default:sub/ba5\x003\x00m3',
         b'',
         ],
        END_OF_MATCH,
        [
         b'branch/default:sub/ba5\x005\x00m5',
         b'branch/default:sub/ba5\x006\x00m6',
         b'branch/default:sub/ba5\x007\x00ba5 content',
         b''
        ],
        END_OF_MATCH,
    ]
    # several matching files, with case insensitivity
    assert len(search(query='ConTent')) == 6 * 2

    # long match requiring several responses
    long_match = search(query='^very', aggregate_splitlines=False)
    assert len(long_match) == 3
    assert long_match[-1] == END_OF_MATCH
    assert len(long_match[0].match_data) == WRITE_BUFFER_SIZE

    # error cases
    with pytest.raises(grpc.RpcError) as exc_info:
        search()
    assert exc_info.value.code() == INVALID_ARGUMENT


@parametrize('rpc_name', ['restore', 'set'])
def test_set_custom_hooks(fixture_with_repo, tmpdir, rpc_name):
    fixture = fixture_with_repo
    wrapper = fixture.repo_wrapper

    if rpc_name == 'restore':
        meth = fixture.restore_custom_hooks
    else:
        meth = fixture.set_custom_hooks

    # this is a valid tar file with the minimal `hgrc` used in these tests
    # the underlying restoration function from py-heptapod is itself tested,
    # and Heptapod backup tests will further exercise this.
    # We only need a minimal set of assertions here.
    tarball_path = TEST_DATA_DIR / 'backup_additional_no_git.tar'
    hgrc_path = wrapper.path / '.hg/hgrc'

    os.unlink(hgrc_path)
    meth(tarball_path)
    assert hgrc_path.read_text()

    # TODO test unknown storage and repo

    # invalid tarball (seems to be what Gitaly returns, to be confirmed
    # by a comparison test)
    invalid_tarball_path = tmpdir / 'invalid.tgz'
    invalid_tarball_path.write("something else")
    with pytest.raises(grpc.RpcError) as exc_info:
        meth(invalid_tarball_path)
    assert exc_info.value.code() == INTERNAL


@parametrize('rpc_name', ['backup', 'get'])
def test_backup_custom_hooks(fixture_with_repo, tmpdir, rpc_name):
    fixture = fixture_with_repo
    if rpc_name == 'backup':
        meth = fixture.backup_custom_hooks
    else:
        meth = fixture.get_custom_hooks

    tarball_path = tmpdir / 'hg_aditional.tgz'
    meth(tarball_path)
    assert tarball_path.exists()


@parametrize('wds_success', ('ok', 'error'))
def test_remove_repository(fixture_with_repo, wds_success):
    fixture = fixture_with_repo
    wrapper, grpc_repo = fixture.repo_wrapper, fixture.grpc_repo
    repo = wrapper.repo
    tmp_dir = wrapper.path.parent / '+hgitaly/tmp'
    wds_root = tmp_dir / 'workdirs' / grpc_repo.relative_path

    if wds_success == 'error':
        # let's sabotage the roster file to trigger an inner exception
        working_dirs_roster_path = repo.vfs.join(WDS_ROSTER_FILE_NAME)
        os.symlink(working_dirs_roster_path, working_dirs_roster_path)
    else:
        with working_directory(wds_root, repo):
            pass

    fixture.remove_repository()
    assert not wrapper.path.exists()

    # no other leftovers alongside the removed repo
    assert os.listdir(fixture.storage_path()) == ['+hgitaly']

    # no leftover in the temporary directory either
    assert not os.listdir(tmp_dir / 'workdirs')

    # unknown storage and repo
    with pytest.raises(grpc.RpcError) as exc_info:
        fixture.remove_repository(grpc_repo=Repository(storage_name='unknown'))
    assert exc_info.value.code() == INVALID_ARGUMENT

    with pytest.raises(grpc.RpcError) as exc_info:
        fixture.remove_repository(
            grpc_repo=Repository(storage_name=grpc_repo.storage_name,
                                 relative_path='no/such/path'))
    assert exc_info.value.code() == NOT_FOUND

    # missing repo argument
    fixture.grpc_repo = None
    with pytest.raises(grpc.RpcError) as exc_info:
        fixture.remove_repository(grpc_repo=None)
    assert exc_info.value.code() == INVALID_ARGUMENT


def test_fetch_bundle(fixture_with_repo, tmpdir):
    fixture = fixture_with_repo
    wrapper = fixture.repo_wrapper

    target_repo_path, target_repo_msg = fixture.additional_repo('target_repo')

    # first call will create the repo
    sha1 = wrapper.commit_file('foo').hex()
    bundle_path = tmpdir / '1.bundle'
    wrapper.command('bundle', as_bytes(bundle_path), all=True)
    fixture.fetch_bundle(bundle_path, target_repo_msg)
    fixture.assert_compare_repos(target_repo_path)

    # subsequent call adds a changeset
    wrapper.commit_file('inc')
    # this seems to be either a minor bug in Mercurial 6.4 or in
    # mercurial-testhelpers: the `tip` pseudo tag is not immediately updated
    # leading to the `assert_compare_repos` failing.
    wrapper.reload()
    bundle2_path = tmpdir / '2.bundle'
    wrapper.command('bundle', as_bytes(bundle2_path), rev=[b'.'], base=[sha1])
    fixture.fetch_bundle(bundle2_path, target_repo_msg, nb_chunks=1)
    fixture.assert_compare_repos(target_repo_path)

    # when storage does not exist
    with pytest.raises(grpc.RpcError) as exc_info:
        fixture.fetch_bundle(bundle_path,
                             Repository(relative_path='new_target_repo',
                                        storage_name='hippoship'))
    assert exc_info.value.code() == INVALID_ARGUMENT
    assert 'No storage named' in exc_info.value.details()

    # missing repository argument
    with pytest.raises(grpc.RpcError) as exc_info:
        fixture.fetch_bundle(bundle_path, None)
    assert exc_info.value.code() == INVALID_ARGUMENT


def test_create_bundle_and_create_repository_from_bundle(fixture_with_repo,
                                                         tmpdir):
    fixture = fixture_with_repo
    grpc_repo = fixture.grpc_repo
    wrapper = fixture.repo_wrapper
    # repo structure:
    #
    #   @ 3 zoo (phase: draft) (amended)
    #   |
    #   | o 1 bar (phase: public) (tag: v1.2.3)
    #   |/
    #   o 0 foo (phase: public)
    #
    ctx0 = wrapper.commit_file('foo')
    sha0 = ctx0.hex()
    sha1 = wrapper.commit_file('bar').hex()
    wrapper.commit_file('zoo', parent=ctx0)
    wrapper.amend_file('zoo')
    wrapper.set_phase('public', [sha0, sha1])
    wrapper.update(sha1)
    wrapper.command('tag', b'v1.2.3', rev=sha1)

    bundle_path = tmpdir / 'my.bundle'
    target_repo_path, target_repo_msg = fixture.additional_repo('target_repo')

    # Bundling and creating new repo from bundle
    fixture.create_bundle(bundle_path)
    fixture.create_repository_from_bundle(bundle_path, target_repo_msg)

    # compare the created repository content with original one
    fixture.assert_compare_repos(target_repo_path)

    # error case of repository creation when target repo already exists
    with pytest.raises(grpc.RpcError) as exc_info:
        fixture.create_repository_from_bundle(bundle_path,
                                              target_repo_msg)
    assert exc_info.value.code() == ALREADY_EXISTS
    details = exc_info.value.details()
    assert 'exists already' in details

    # when storage name doesn't exists
    with pytest.raises(grpc.RpcError) as exc_info:
        fixture.create_repository_from_bundle(
            bundle_path,
            Repository(relative_path='new_target_repo',
                       storage_name='hippoship')
        )
    assert exc_info.value.code() == INVALID_ARGUMENT
    assert 'no such storage' in exc_info.value.details()

    # test errors while creating repo itself
    # Because the main code refuses existing paths, we make a parent dir that
    # is a broken symlink (which points to itself). This way the repo path
    # does not exist *and* its creation fails.
    br_dir = Path("my_broken_symlink")
    br_path = fixture.repo_path(br_dir)
    br_path.symlink_to(br_path)
    trepo_msg = Repository(relative_path=str(br_dir / 'repo'),
                           storage_name=grpc_repo.storage_name)
    with pytest.raises(grpc.RpcError) as exc_info:
        fixture.create_repository_from_bundle(bundle_path, trepo_msg)
    assert exc_info.value.code() == INTERNAL
    assert 'Too many levels of symbolic links' in exc_info.value.details()

    # in case of error after repo creation, the repo must be removed
    trepo_msg = Repository(relative_path='bundle-fail',
                           storage_name=grpc_repo.storage_name)
    broken_bundle_path = tmpdir / 'broken-bundle'
    broken_bundle_path.write("Obviously garbage")
    with pytest.raises(grpc.RpcError) as exc_info:
        fixture.create_repository_from_bundle(broken_bundle_path, trepo_msg)
    assert exc_info.value.code() == INTERNAL
    assert "bundle" in exc_info.value.details()
    assert not fixture.repo_path('bundle-fail').exists()


def test_create_bundle_from_ref_list(fixture_with_repo, tmpdir):
    fixture = fixture_with_repo
    wrapper, grpc_repo = fixture.repo_wrapper, fixture.grpc_repo

    # repo structure:
    #
    #   @ 4 zoo (branch:toys) (phase: public) (amended)
    #   |
    #   | o 2 added tag (branch: animal) (phase: draft)
    #   | |
    #   | o 1 bar (tag: v1.2.3)
    #   |/
    #   o 0 foo (branch: default)
    #
    ctx0 = wrapper.commit_file('foo')
    sha0 = ctx0.hex()
    sha1 = wrapper.commit_file('bar', branch='animal').hex()
    wrapper.command('tag', b'v1.2.3')
    obs_sha = wrapper.commit_file('zoo', parent=ctx0, branch='toys').hex()
    sha4 = wrapper.amend_file('zoo').hex()
    wrapper.set_phase('public', [sha0, sha4])

    def target_repo_msg(bundle_name):
        return Repository(relative_path='%s_repo' % (bundle_name),
                          storage_name=grpc_repo.storage_name)

    def target_repo(bundle_name):
        return fixture.make_repo_wrapper(
            '%s_repo' % bundle_name
        ).repo.unfiltered()

    def assert_compare_created_repository_from_bundle(
            bundle_name, exp_branches, exp_tags, exp_cids):
        repo = target_repo(bundle_name)

        # assert branches
        branch_prefix = b'refs/heads/branch/'
        branches = set(
            [branch_prefix + br[0] for br in repo.branchmap().iterbranches()])
        assert branches == set(exp_branches)

        # assert tags
        tag_prefix = b'refs/tags/'
        hg_tags = set(map(lambda t: tag_prefix + t, repo.tags().keys()))
        hg_tags.remove(b'refs/tags/tip')
        assert hg_tags == set(exp_tags)

        # assert included revs
        allrevs = repo.revs('all()')
        revs = set(map(lambda x: repo[x].hex(), allrevs))
        assert revs == set(exp_cids)

    repo = wrapper.repo.unfiltered()
    all_cids = [repo[0].hex(), repo[1].hex(), repo[2].hex(), repo[4].hex()]
    all_cids_unfiltered = all_cids + [obs_sha]

    # test with all refs
    allrefs_bundle = 'all_refs_bundle'
    allrefs_bundle_path = tmpdir / allrefs_bundle
    branch_list = [b'refs/heads/branch/animal', b'refs/heads/branch/toys',
                   b'refs/heads/branch/default']
    tag_list = [b'refs/tags/v1.2.3']

    fixture.create_bundle_from_ref_list(allrefs_bundle_path,
                                        branch_list + tag_list)
    fixture.create_repository_from_bundle(allrefs_bundle_path,
                                          target_repo_msg(allrefs_bundle))

    # test successful repo creation from bundle
    assert_compare_created_repository_from_bundle(
        allrefs_bundle,
        exp_branches=branch_list,
        exp_tags=tag_list,
        exp_cids=all_cids,
    )

    # test with the `ALL` pseudo ref
    full_bundle = 'full_bundle'
    full_bundle_path = tmpdir / full_bundle
    fixture.create_bundle_from_ref_list(full_bundle_path,
                                        [b'ALL'])
    fixture.create_repository_from_bundle(full_bundle_path,
                                          target_repo_msg(full_bundle))
    assert_compare_created_repository_from_bundle(
        full_bundle,
        exp_branches=branch_list,
        exp_tags=tag_list,
        exp_cids=all_cids_unfiltered,
    )

    # test with some full ref paths
    someref_bundle = 'some_refs_bundle'
    someref_path = tmpdir / someref_bundle
    ref_list = [b'refs/heads/branch/default', b'refs/heads/branch/toys']

    fixture.create_bundle_from_ref_list(someref_path, ref_list)
    fixture.create_repository_from_bundle(someref_path,
                                          target_repo_msg(someref_bundle))

    # test successful repo creation from bundle
    assert_compare_created_repository_from_bundle(
        someref_bundle,
        exp_branches=ref_list,
        exp_tags=[],
        exp_cids=[sha0, sha4],
    )

    # test error case: when repository is absent in request obj
    with pytest.raises(grpc.RpcError) as exc_info_hg:
        fixture.create_bundle_from_ref_list(
            tmpdir / 'temp_bname', [b'refs/heads/branch/default'],
            without_repository=True)
    assert exc_info_hg.value.code() == INVALID_ARGUMENT
    assert exc_info_hg.value.details() == 'repository not set'

    # test incrementality with negative revspecs

    # first bundle won't have the obsolete changeset, since it is not
    # an ancestor of `refs/heads/branch/toys`.
    inc_bundle_1, inc_bundle2 = 'bundle_inc.1', 'bundle_inc.2'
    inc_path_1, inc_path_2 = tmpdir / inc_bundle_1, tmpdir / inc_bundle2
    fixture.create_bundle_from_ref_list(inc_path_1, [b'branch/toys', sha1])
    fixture.create_bundle_from_ref_list(inc_path_2,
                                        [b'branch/animal', b'^' + sha1])

    # loading the two bundles is equivalent to loading the full backup
    # without the obsolete changeset
    target_grpc_repo = target_repo_msg('incremental')
    fixture.create_repository_from_bundle(inc_path_1, target_grpc_repo)
    fixture.fetch_bundle(inc_path_2, target_grpc_repo)
    assert_compare_created_repository_from_bundle(
        'incremental',
        exp_branches=branch_list,
        exp_tags=tag_list,
        exp_cids=all_cids,
    )

    # Same by including obsolete changesets, thanks to the ALL pseudo-ref
    # Also adding a phase change, so that we can test if it gets bundled.
    wrapper.set_phase('public', [sha1])
    wrapper.update('toys')
    obs_sha2 = wrapper.commit_file('zoo', message='obs2').hex()
    sha5 = wrapper.amend_file('zoo').hex()
    inc_bundle_3 = 'bundle_inc.3'
    inc_path_3 = tmpdir / inc_bundle_3
    exclude_previous = [b'^' + sha for sha in all_cids]
    fixture.create_bundle_from_ref_list(inc_path_3,
                                        [b'ALL'] + exclude_previous)
    inc_cids = set(list_bundle_contents(inc_path_3)['nodes'])

    # caveat: obs_sha is present in incremental bundle, because it is
    # not an ancestor of previously backed up changesets.
    assert inc_cids == {obs_sha2, sha5, obs_sha}

    fixture.fetch_bundle(inc_path_3, target_grpc_repo)
    all_cids_unfiltered.extend((obs_sha2, sha5))
    assert_compare_created_repository_from_bundle(
        'incremental',
        exp_branches=branch_list,
        exp_tags=tag_list,
        exp_cids=all_cids_unfiltered,
    )
    unbundled_repo = target_repo('incremental')

    # confirmation that phase change of previously backed up changeset
    # is not in the incremental bundle
    # TODO do better and assert phases.public
    assert unbundled_repo[sha1].phase() == phases.draft


def test_find_license(fixture_with_repo):
    find_license = fixture_with_repo.find_license
    wrapper = fixture_with_repo.repo_wrapper

    # empty repo
    assert find_license() == FindLicenseResponse()

    wrapper.commit_file('COPYING', message="lic",
                        content=license_content('GPL-2'))
    set_default_gitlab_branch(wrapper.repo, b'branch/default')
    resp = find_license()
    # this result is wrong in the absolute but it is really what
    # go-license-detector and hence Gitaly returns (check comparison tests)
    assert resp.license_short_name == 'gpl-2.0+'
    assert resp.license_path == 'COPYING'

    wrapper.commit_file('COPYING', content="Some garbage")
    resp = find_license()
    assert resp.license_path == 'COPYING'
    assert resp.license_short_name == 'other'

    # test with a subrepo, see heptapod#1287
    nested_path = wrapper.path / 'nested'
    nested = RepoWrapper.init(nested_path)
    nested.commit_file("bar", content="in nested")
    (wrapper.path / '.hgsub').write_bytes(b"nested = nested\n")
    wrapper.command(b'add', subrepos=True)
    wrapper.command(b'commit', subrepos=True, message=b"invalid")
    assert find_license() == FindLicenseResponse()


def test_object_format(fixture_with_repo):
    fixture = fixture_with_repo
    assert (fixture.object_format().format
            == ObjectFormat.OBJECT_FORMAT_UNSPECIFIED)



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later

from ..oid import (
    tree_oid,
    blob_oid,
    extract_blob_oid,
    extract_tree_oid,
)


def test_tree_oid():
    chgs_id = "01be23ef" * 5
    path = b"s\xfcb/dir"
    repo = None  # actual value not needed for the time being
    oid = tree_oid(repo, chgs_id, path)
    assert extract_tree_oid(repo, oid) == (chgs_id, path)

    oid = tree_oid(repo, chgs_id, b'')
    assert oid != chgs_id
    assert extract_tree_oid(repo, oid) == (chgs_id, b'')


def test_blob_oid():
    chgs_id = "56de78ad" * 5
    path = b"rang\xe9"
    repo = None  # actual value not needed for the time being
    oid = blob_oid(repo, chgs_id, path)
    assert extract_blob_oid(repo, oid) == (chgs_id, path)


def test_commit_oid():
    chgs_id = "56de78ad" * 5
    repo = None  # actual value not needed for the time being
    assert extract_tree_oid(repo, chgs_id) == (chgs_id, None)



# Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from hashlib import sha256
import json
import re
import subprocess

from grpc import StatusCode
from importlib_resources import files


def read_linguist_languages(path):
    res = {}
    with path.open() as langf:
        for name, details in json.load(langf).items():
            color = details.get('color')
            if color is not None:
                res[name] = color
    return res


LANGUAGE_COLORS = read_linguist_languages(
    files(__name__).joinpath('languages.json'))

TOKEI_VERSION_RX = re.compile(r'tokei (\d+)\.(\d+)\.(\d+)(\s|$)')
TOKEI_VERSIONS = {}


def language_color(lang_name):
    """Perform the same defaulting as in Gitaly.

    This is not thread-safe, because it actually updates
    :data:`LANGUAGE_COLORS` due to defaulting in case the detected language
    does not have a color there. That is not a problem: HGitaly is not
    multithreaded (or rather uses a unique service thread at a given time,
    out of the pool still created by grpcio).

    As of this writing, go-enry v2.8.3 (already used by Gitaly to return the
    color) defaults to `#cccccc`:

      // GetColor returns a HTML color code of a given language.
      func GetColor(language string) string {
        if color, ok := data.LanguagesColor[language]; ok {
            return color
        }

        if color, ok := data.LanguagesColor[GetLanguageGroup(language)]; ok {
            return color
        }

        return "#cccccc"
      }

    Gitaly then concludes with (from internal/gitaly/linguist/linguist.go) ::

      // Color returns the color Linguist has assigned to language.
      func Color(language string) string {
        if color := enry.GetColor(language); color != "#cccccc" {
            return color
        }

        colorSha := sha256.Sum256([]byte(language))
        return fmt.Sprintf("#%x", colorSha[0:3])
      }
    """
    color = LANGUAGE_COLORS.get(lang_name)
    if color is not None:
        return color

    color_sha = sha256(lang_name.encode('utf-8')).hexdigest()
    color = '#' + color_sha[:6]
    LANGUAGE_COLORS[lang_name] = color
    return color


class TokeiVersionParseError(RuntimeError):
    """Indicates that `tokei --version` could not be parsed.

    It is possible that the executable is just not Tokei
    """


def parse_tokei_version(out):
    """Parse the given ``out``, expected to be output of ``tokei --version``

    :returns: a triple of integers
    :raises: :class:TokeiVersionParseError

    Currently tested versions::

      >>> parse_tokei_version("tokei 12.0.4")
      (12, 0, 4)
      >>> parse_tokei_version("tokei 12.1.2")
      (12, 1, 2)

    Errors::

      >>> try: parse_tokei_version("Unrelated")
      ... except Exception as exc: print(repr(exc))
      TokeiVersionParseError('Unrelated')
    """
    match = TOKEI_VERSION_RX.search(out)
    if match is None:
        raise TokeiVersionParseError(out)
    return tuple(int(match.group(i)) for i in (1, 2, 3))


def tokei_version(bin_path):
    version = TOKEI_VERSIONS.get(bin_path)
    if version is not None:
        return version

    tokei = subprocess.run((bin_path, '--version'),
                           encoding='utf-8',
                           check=True,
                           capture_output=True
                           )
    version = parse_tokei_version(tokei.stdout.strip())
    TOKEI_VERSIONS[bin_path] = version
    return version


def language_stats(bin_path, path, context):
    """Return the full language stats working directory at given path.

    :param str bin_path: path to the Tokei executable
    :param path: the directory to analyze as a :class:`pathlib.Path` instance.
    """
    try:
        version = tokei_version(bin_path)  # may invoke Tokei
        options = []
        if version >= (12, 1):
            options.append('--compact')

        cmd = [bin_path]
        cmd.extend(options)
        cmd.extend(('-o', 'json', path))
        tokei = subprocess.run(
            cmd,
            encoding='utf-8',
            check=True,
            capture_output=True,
        )
    except subprocess.CalledProcessError as exc:
        # in Gitaly 15.5, the resulting error code is 'INTERNAL', with
        # details 'language stats: waiting for linguist: exit status %d'
        context.abort(
            StatusCode.INTERNAL,
            "CommitLanguages: tokei error status code=%d stderr=%r" % (
                exc.returncode, exc.stderr))
    except FileNotFoundError:
        # in Gitaly 15.5, if `gitaly-linguist` cannot be found, the resulting
        # error code is `INTERNAL`, with details 'language stats: waiting
        # for linguist: exit status 127' (same as rewrapping in a shell)
        context.abort(
            StatusCode.INTERNAL,
            "language stats: tokei executable '%s' not found" % bin_path)
    except PermissionError:
        # Gitaly 15.5 behaves similarly in this case as in the not found case,
        # only with exit status 126 in details (same as in a shell, again)
        context.abort(
            StatusCode.INTERNAL,
            "language stats: permission denied to run "
            "tokei executable '%s'" % bin_path)
    except TokeiVersionParseError as exc:
        context.abort(
            StatusCode.INTERNAL,
            "language stats: could not parse Tokei version for executable "
            "'%s' (got %r). Is it even Tokei?" % (bin_path, exc.args[0]))

    # overriding variable also to free buffer immediately, as it can be
    # fairly big, e.g. 10MB for heptapod-rails
    tokei = json.loads(tokei.stdout)

    stats = {}
    total = 0

    for lang_name, details in tokei.items():
        # We can't use the `Total` entry because the ratios we have
        # to produce are expressed in bytes, not lines.
        if lang_name == 'Total':
            continue

        per_file = details['reports']
        total_bytes = 0
        for report in per_file:
            total_bytes += (path / report['name']).stat().st_size
        stats[lang_name] = dict(files=len(per_file),
                                total_bytes=total_bytes,
                                )
        total += total_bytes
    return stats, total



import re

from mercurial import (
    error,
    registrar,
    revsetlang,
)
from mercurial.i18n import _
from mercurial.utils import (
    stringutil,
)

getstring = revsetlang.getstring

revsetpredicate = registrar.revsetpredicate()


@revsetpredicate(b'hpd_dsc_rx', weight=10)
def hpd_dsc_rx(repo, subset, x, re_flags=0):
    """Variant of the core `grep` revset suitable for `commit_message_patterns`

    It matches changeset descriptions only. The same function can be used
    for variants, thanks to :param re_flags:

    The name is short for: "HePtapoD DeSCription Regular eXpression", with
    the ``hpd_`` prefix to indicate that it is Heptapod specific.

    Notes:
    - as of Mercurial 6.3, dashes wouldn't work in predicate names,
      leading to the obscure "unknown revision 'hpd'" error message.
    - the ``re.NOFLAG`` constant was introduced in Python 3.11, which is not
      yet supported by HGitaly.
    """
    try:
        rx = re.compile(getstring(x, _(b"hpd_dsc_rx requires a string")),
                        flags=re_flags)
    except re.error as e:
        raise error.ParseError(
            _(b'invalid regexp pattern: %s') % stringutil.forcebytestr(e)
        )

    def matches(x):
        return rx.search(repo[x].description()) is not None

    return subset.filter(matches, condrepr=(b'<hpd_dsc_rx %r>', rx.pattern))


@revsetpredicate(b'hpd_dsc_irx', weight=10)
def hpd_dsc_irx(repo, subset, x):
    """Case-insensitive variant of hpd_dsc_rx"""
    return hpd_dsc_rx(repo, subset, x, re_flags=re.IGNORECASE)



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import multiprocessing
import pytest
import socket

import grpc

from hgitaly.testing.grpc import wait_server_accepts_connection
from .. import mono as server


def test_exceptions():
    storages = dict(default='/some/path')

    # unterminated IPv6 address, one of the very few cases of unparseable URL
    url = 'tcp://[::'
    with pytest.raises(server.InvalidUrl) as exc_info:
        server.init([url], storages)
    assert exc_info.value.args == (url, 'Invalid IPv6 URL')

    # unsupported scheme
    url = 'exotic://localhost:1234'
    with pytest.raises(server.UnsupportedUrlScheme) as exc_info:
        server.init([url], storages)
    assert exc_info.value.args == ('exotic', )

    # bind error
    url = 'tcp://unresolvable-or-youre-kidding-me:0'
    with pytest.raises(server.BindError) as exc_info:
        server.init([url], storages)
    assert exc_info.value.args == (url, )

    # bind error with fixed port
    url = 'tcp://unresolvable-or-youre-kidding-me:1234'
    with pytest.raises(server.BindError) as exc_info:
        server.init([url], storages)
    assert exc_info.value.args == (url, )

    # bind error for non available port
    sock = None
    try:
        sock = socket.socket(socket.AF_INET)
        sock.bind(('127.0.0.1', 0))
        url = 'tcp://%s:%d' % sock.getsockname()
        with pytest.raises(server.BindError) as exc_info:
            server.init([url], storages)
    finally:
        assert sock is not None
        sock.close()
    assert exc_info.value.args == (url, )


def test_init_tcp():
    storages = dict(default='/some/path')
    server_instance = server.init(['tcp://localhost:0'], storages)
    try:
        server_instance.start()
    finally:
        server_instance.stop(None)
        server_instance.wait_for_termination()


def test_init_unix(tmpdir):
    storages = dict(default='/some/path')
    # 'unix://PATH' requires PATH to be absolute, whereas 'unix:PATH' does not.
    # currently, str(tmpdir) seems to be always absolute, but let's not depend
    # on that.
    url = 'unix:%s' % tmpdir.join('hgitaly.socket')
    server_instance = server.init([url], storages)
    try:
        server_instance.start()
    finally:
        server_instance.stop(None)
        server_instance.wait_for_termination()


def test_server_process(tmpdir):
    socket_path = tmpdir.join('hgitaly.socket')
    repos_root = tmpdir.join('repos').ensure(dir=True)
    listen_url = 'unix://' + str(socket_path)
    p = multiprocessing.Process(
        target=server.server_process,
        args=(0, [listen_url],
              dict(default=str(repos_root).encode())),
        kwargs=dict(pytest_cov=True),
    )

    p.start()
    with grpc.insecure_channel(listen_url) as channel:
        wait_server_accepts_connection(channel)

    p.terminate()

    p.join(2)  # long timeout, we will wait only if test fails
    try:
        assert p.exitcode is not None
    except Exception:  # pragma no cover
        # let's avoid hanging on failed test
        p.kill()
        p.join()



# Copyright 2021 Sushil Khanchi <sushilkhanchi97@gmail.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import grpc
from mercurial.pycompat import sysstr
import pytest
from hgitaly.git import (
    EMPTY_TREE_OID,
    OBJECT_MODE_DOES_NOT_EXIST,
    OBJECT_MODE_EXECUTABLE,
    OBJECT_MODE_NON_EXECUTABLE,
)
from hgitaly.oid import (
    tree_oid,
)
from hgitaly.stub.diff_pb2 import (
    CommitDeltaRequest,
    CommitDiffRequest,
    ChangedPaths,
    DiffStatsRequest,
    GetPatchIDRequest,
    RawDiffRequest,
    RawPatchRequest,
    FindChangedPathsRequest,
)
from mercurial import (
    node,
)
from hgitaly.stub.diff_pb2_grpc import DiffServiceStub

from .fixture import ServiceFixture

StatusCode = grpc.StatusCode


class DiffFixture(ServiceFixture):

    stub_cls = DiffServiceStub

    def raw_method(self, stub_meth, req_cls, left_sha, right_sha):
        request = req_cls(
            repository=self.grpc_repo,
            left_commit_id=left_sha,
            right_commit_id=right_sha,
        )
        return b''.join(resp.data for resp in stub_meth(request))

    def raw_diff(self, left_sha, right_sha):
        return self.raw_method(self.stub.RawDiff,
                               RawDiffRequest,
                               left_sha, right_sha)

    def raw_patch(self, left_sha, right_sha):
        return self.raw_method(self.stub.RawPatch,
                               RawPatchRequest,
                               left_sha, right_sha)

    def commit_ids_method(self, stub_meth, req_cls,
                          left_cid, right_cid,
                          flatten_attr=None,
                          **kwargs):
        """Common caller for commit_* methods and similar.

        :param flatten_attr: if provided, the responses are flattened to
           a single list, from sublists provided in this attribute. Other
           wise simple conversion to list of the streaming request is
           performed.
        """
        response = stub_meth(req_cls(repository=self.grpc_repo,
                                     left_commit_id=left_cid,
                                     right_commit_id=right_cid,
                                     **kwargs))
        if flatten_attr is not None:
            return [item for r in response
                    for item in getattr(r, flatten_attr)]
        return list(response)

    def commit_diff(self, left_cid, right_cid, **kwargs):
        return self.commit_ids_method(self.stub.CommitDiff,
                                      CommitDiffRequest,
                                      left_cid, right_cid,
                                      **kwargs)

    def get_patch_id(self, old_revision, new_revision, **kwargs):
        kwargs.setdefault('repository', self.grpc_repo)
        return self.stub.GetPatchID(GetPatchIDRequest(
            old_revision=old_revision,
            new_revision=new_revision,
            **kwargs)).patch_id

    def commit_delta(self, left_cid, right_cid, **kwargs):
        return self.commit_ids_method(self.stub.CommitDelta,
                                      CommitDeltaRequest,
                                      left_cid, right_cid,
                                      flatten_attr='deltas',
                                      **kwargs)

    def diff_stats(self, left_cid, right_cid, **kwargs):
        return self.commit_ids_method(self.stub.DiffStats,
                                      DiffStatsRequest,
                                      left_cid, right_cid,
                                      flatten_attr='stats',
                                      **kwargs)

    def find_changed_paths(self, **kwargs):
        """Run FindChangedPaths and present the results by file path.

        :return: a :class:`dict` whose keys are file paths (bytes) and
          values are lists of (status change, permission change).

        Permission changes are pairs ``(old_perm, new_perm)`` if this is
        an actual change, and ``None`` otherwise.
        """
        by_file = {}
        for resp in self.stub.FindChangedPaths(FindChangedPathsRequest(
                repository=self.grpc_repo, **kwargs)):
            for changed in resp.paths:
                if changed.old_mode != changed.new_mode:
                    perm_change = (changed.old_mode, changed.new_mode)
                else:
                    perm_change = None

                by_file.setdefault(changed.path, []).append(
                    (changed.status, perm_change))
        return by_file

    def find_changed_paths_commits(self, commits, compare_to=()):
        """Wrap FindChangedPaths used with CommitRequest.

        :param compare_to: if given, will be used in all requests, hoping
          we won't need more flexibility
        """
        CommitRequest = FindChangedPathsRequest.Request.CommitRequest
        return self.find_changed_paths(
            requests=[
                FindChangedPathsRequest.Request(
                    commit_request=CommitRequest(
                        commit_revision=c,
                        parent_commit_revisions=compare_to))
                for c in commits],
        )

    def find_changed_paths_tree(self, left_oid, right_oid):
        TreeRequest = FindChangedPathsRequest.Request.TreeRequest
        return self.find_changed_paths(
            requests=[
                FindChangedPathsRequest.Request(
                    tree_request=TreeRequest(left_tree_revision=left_oid,
                                             right_tree_revision=right_oid,
                                             ))
                ]
        )


@pytest.fixture
def diff_fixture(grpc_channel, server_repos_root):
    with DiffFixture(grpc_channel, server_repos_root) as fixture:
        yield fixture


def test_raw_diff(diff_fixture):
    raw_diff = diff_fixture.raw_diff
    wrapper = diff_fixture.repo_wrapper

    ctx0 = wrapper.commit_file('foo', content="I am oof\n",
                               message=b'added foo')
    ctx1 = wrapper.commit_file('foo', content="I am foo\n",
                               message=b'changes foo')
    wrapper.command(b'mv', wrapper.repo.root + b'/foo',
                    wrapper.repo.root + b'/zoo')
    wrapper.command(b'ci', message=b"rename foo to zoo")
    ctx2 = wrapper.repo[b'.']
    sha0, sha1, sha2 = ctx0.hex(), ctx1.hex(), ctx2.hex()

    # case 1: actual test
    resp = raw_diff(sha0, sha1)
    respheader = (
        b'diff --git a/foo b/foo\n'
    )
    resphunk = (
        b'--- a/foo\n'
        b'+++ b/foo\n'
        b'@@ -1,1 +1,1 @@\n'
        b'-I am oof\n'
        b'+I am foo\n'
    )
    assert resp.startswith(respheader) and resp.endswith(resphunk)

    # case 2: with null node
    resp = raw_diff(node.nullhex, sha0)
    respheader = (
        b'diff --git a/foo b/foo\n'
    )
    resphunk = (
        b'--- /dev/null\n'
        b'+++ b/foo\n'
        b'@@ -0,0 +1,1 @@\n'
        b'+I am oof\n'
    )
    assert resp.startswith(respheader) and resp.endswith(resphunk)

    # case 2bis: with null left node, expressed as empty tree
    # this is really used by the Rails app.
    resp = raw_diff(EMPTY_TREE_OID, sha0)
    assert resp.startswith(respheader) and resp.endswith(resphunk)

    # case 2ter: with null right node, expressed as empty tree
    # this is for completeness
    resp = raw_diff(sha0, EMPTY_TREE_OID)
    resphunk = (
        b'--- a/foo\n'
        b'+++ /dev/null\n'
        b'@@ -1,1 +0,0 @@\n'
        b'-I am oof\n'
    )
    assert resp.startswith(respheader) and resp.endswith(resphunk)

    # case 3: with file renaming
    resp = raw_diff(sha1, sha2)
    assert resp == (
        b'diff --git a/foo b/zoo\n'
        b'similarity index 100%\n'
        b'rename from foo\n'
        b'rename to zoo\n'
    )

    # case 4: when commit_id does not correspond to a commit
    sha_not_exists = b'deadnode' * 5
    # varient 1
    with pytest.raises(grpc.RpcError) as exc_info:
        raw_diff(sha0, sha_not_exists)
    assert exc_info.value.code() == grpc.StatusCode.INTERNAL
    assert 'exit status 128' in exc_info.value.details()
    # varient 2
    with pytest.raises(grpc.RpcError) as exc_info:
        raw_diff(sha_not_exists, sha0)
    assert exc_info.value.code() == grpc.StatusCode.INTERNAL
    assert 'exit status 128' in exc_info.value.details()


def test_raw_patch(diff_fixture):
    raw_patch = diff_fixture.raw_patch

    # prepare repo as:
    #
    #   @    3 merge with stable
    #   |\
    #   | o  2 added bar (branch: stable)
    #   | |
    #   o |  1 changes foo (topic: feature)
    #   |/
    #   o  0 added foo
    #
    #
    wrapper = diff_fixture.repo_wrapper
    dayoffset = 86400  # seconds in 24 hours
    ctx0 = wrapper.commit_file('foo', content="I am oof\n",
                               message=b'added foo', user=b'testuser',
                               utc_timestamp=dayoffset)
    ctx1 = wrapper.commit_file('foo', content="I am foo\n", topic=b'feature',
                               message=b'changes foo', user=b'testuser',
                               utc_timestamp=dayoffset*2)
    ctx2 = wrapper.commit_file('bar', content="I am bar\n",
                               message=b'added bar', user=b'testuser',
                               utc_timestamp=dayoffset*3, parent=ctx0,
                               branch=b'stable')
    wrapper.update(ctx1.rev())
    ctx3 = wrapper.merge_commit(ctx2, user=b'testuser',
                                utc_timestamp=dayoffset*4,
                                message=b'merge with stable')
    sha0, sha2, sha3 = ctx0.hex(), ctx2.hex(), ctx3.hex()

    # with null revision
    null_node = b"00000" * 5
    assert raw_patch(null_node, sha0) == (
        b'# HG changeset patch\n'

        b'# User testuser\n'
        b'# Date 86400 0\n'
        b'#      Fri Jan 02 00:00:00 1970 +0000\n'
        b'# Node ID f1a2b5b072f5e59abd43ed6982ab428a6149eda8\n'
        b'# Parent  0000000000000000000000000000000000000000\n'
        b'added foo\n'
        b'\n'
        b'diff --git a/foo b/foo\n'
        b'new file mode 100644\n'
        b'--- /dev/null\n'
        b'+++ b/foo\n'
        b'@@ -0,0 +1,1 @@\n'
        b'+I am oof\n'
    )
    # with merge commit
    assert raw_patch(sha2, sha3) == (
        b'# HG changeset patch\n'
        b'# User testuser\n'
        b'# Date 172800 0\n'
        b'#      Sat Jan 03 00:00:00 1970 +0000\n'
        b'# Node ID 0ae85a0d494d9197fd2bf8347d7fff997576f25a\n'
        b'# Parent  f1a2b5b072f5e59abd43ed6982ab428a6149eda8\n'
        b'# EXP-Topic feature\n'
        b'changes foo\n'
        b'\n'
        b'diff --git a/foo b/foo\n'
        b'--- a/foo\n'
        b'+++ b/foo\n'
        b'@@ -1,1 +1,1 @@\n'
        b'-I am oof\n'
        b'+I am foo\n'
        b'# HG changeset patch\n'
        b'# User testuser\n'
        b'# Date 345600 0\n'
        b'#      Mon Jan 05 00:00:00 1970 +0000\n'
        b'# Node ID 2215a964a3245ee4e7c3906f076b14977152a1df\n'
        b'# Parent  0ae85a0d494d9197fd2bf8347d7fff997576f25a\n'
        b'# Parent  c4fa3ef1fc8ba157ed8c26584c13492583bf17e9\n'
        b'# EXP-Topic feature\n'
        b'merge with stable\n'
        b'\n'
        b'diff --git a/bar b/bar\n'
        b'new file mode 100644\n'
        b'--- /dev/null\n'
        b'+++ b/bar\n'
        b'@@ -0,0 +1,1 @@\n'
        b'+I am bar\n'
    )
    # with different branch
    assert raw_patch(sha0, sha2) == (
        b'# HG changeset patch\n'
        b'# User testuser\n'
        b'# Date 259200 0\n'
        b'#      Sun Jan 04 00:00:00 1970 +0000\n'
        b'# Branch stable\n'
        b'# Node ID c4fa3ef1fc8ba157ed8c26584c13492583bf17e9\n'
        b'# Parent  f1a2b5b072f5e59abd43ed6982ab428a6149eda8\n'
        b'added bar\n'
        b'\n'
        b'diff --git a/bar b/bar\n'
        b'new file mode 100644\n'
        b'--- /dev/null\n'
        b'+++ b/bar\n'
        b'@@ -0,0 +1,1 @@\n'
        b'+I am bar\n'
    )
    # when commit_id does not correspond to a commit
    sha_not_exists = b'deadnode' * 5
    # varient 1
    with pytest.raises(grpc.RpcError) as exc_info:
        raw_patch(sha0, sha_not_exists)
    assert exc_info.value.code() == grpc.StatusCode.INTERNAL
    assert 'exit status 128' in exc_info.value.details()
    # varient 2
    with pytest.raises(grpc.RpcError) as exc_info:
        raw_patch(sha_not_exists, sha0)
    assert exc_info.value.code() == grpc.StatusCode.INTERNAL
    assert 'exit status 128' in exc_info.value.details()


def test_commit_diff(diff_fixture):
    commit_diff = diff_fixture.commit_diff

    wrapper = diff_fixture.repo_wrapper
    ctx0 = wrapper.commit_file('bar', content="I am in\nrab\n",
                               message="Add bar")
    ctx1 = wrapper.commit_file('bar', content="I am in\nbar\n",
                               message="Changes bar")
    wrapper.command(b'mv', wrapper.repo.root + b'/bar',
                    wrapper.repo.root + b'/zar')
    ctx2 = wrapper.commit([b'bar', b'zar'], message="Rename bar to zar")
    ctx3 = wrapper.commit_file('zoo', content="I am in\nzoo\n",
                               message="Added zoo")
    ctx4 = wrapper.commit_file('zoo', content="I  am in\n zoo\n")
    # Repo structure:
    #
    # @  4 Whitespace changes
    # |
    # o  3 Added zoo
    # |
    # o  2 Rename bar to zar
    # |
    # o  1 Changes bar
    # |
    # o  0 Add bar
    #

    # case 1: when a file renamed
    resp = commit_diff(left_cid=ctx1.hex(), right_cid=ctx2.hex())
    assert len(resp) == 1
    resp = resp[0]
    assert resp.from_path == b'bar'
    assert resp.to_path == b'zar'
    assert resp.raw_patch_data == b''
    assert resp.old_mode == resp.new_mode

    # case 2: when a new file added
    resp = commit_diff(left_cid=ctx2.hex(), right_cid=ctx3.hex())
    assert len(resp) == 1
    resp = resp[0]
    assert resp.from_path == resp.to_path
    assert resp.raw_patch_data == b'@@ -0,0 +1,2 @@\n+I am in\n+zoo\n'
    assert resp.old_mode == 0
    assert resp.new_mode == 0o100644

    # case 3: test with enforce_limits
    # Note: For thorough testing, look at comparison tests
    resp = commit_diff(
        left_cid=ctx2.hex(),
        right_cid=ctx3.hex(),
        enforce_limits=True,
        max_files=10,
        max_bytes=100,
        max_lines=1,
    )
    assert len(resp) == 1
    assert resp[0].overflow_marker

    # case 4: test with collapse_diffs
    # Note: For thorough testing, look at comparison tests
    resp = commit_diff(
        left_cid=ctx2.hex(),
        right_cid=ctx3.hex(),
        collapse_diffs=True,
        safe_max_files=10,
        safe_max_bytes=100,
        safe_max_lines=1,
    )
    assert len(resp) == 1
    assert resp[0].collapsed

    # case 5: test with paths
    resp = commit_diff(left_cid=ctx0.hex(), right_cid=ctx3.hex(),
                       paths=[b'zoo'])
    assert len(resp) == 1
    resp = resp[0]
    assert resp.from_path == resp.to_path
    assert resp.from_path == b'zoo'
    assert resp.raw_patch_data == b'@@ -0,0 +1,2 @@\n+I am in\n+zoo\n'

    # case 6: when commit_id does not correspond to a commit
    sha_not_exists = b'deadnode' * 5
    with pytest.raises(grpc.RpcError) as exc_info:
        commit_diff(sha_not_exists, ctx3.hex())
    assert exc_info.value.code() == grpc.StatusCode.FAILED_PRECONDITION

    # cases 7 and 8: ignoring whitespace
    WHITESPACE_IGNORE = CommitDiffRequest.WHITESPACE_CHANGES_IGNORE
    WHITESPACE_IGNORE_ALL = CommitDiffRequest.WHITESPACE_CHANGES_IGNORE_ALL
    resp = commit_diff(left_cid=ctx3.hex(), right_cid=ctx4.hex(),
                       whitespace_changes=WHITESPACE_IGNORE)
    assert len(resp) == 1
    resp = resp[0]
    assert resp.from_path == resp.to_path
    assert resp.raw_patch_data == (b'@@ -1,2 +1,2 @@\n'
                                   b' I am in\n'
                                   b'-zoo\n+ zoo\n')

    resp = commit_diff(left_cid=ctx3.hex(), right_cid=ctx4.hex(),
                       whitespace_changes=WHITESPACE_IGNORE_ALL)
    assert len(resp) == 0


def test_commit_delta(diff_fixture):
    wrapper = diff_fixture.repo_wrapper
    commit_delta = diff_fixture.commit_delta

    ctx0 = wrapper.commit_file('bar', content="I am in\nrab\n",
                               message="Add bar")
    ctx1 = wrapper.commit_file('bar', content="I am in\nbar\n",
                               message="Changes bar")
    wrapper.command(b'mv', wrapper.repo.root + b'/bar',
                    wrapper.repo.root + b'/zar')
    ctx2 = wrapper.commit([b'bar', b'zar'], message="Rename bar to zar")
    ctx3 = wrapper.commit_file('zoo', content="I am in\nzoo\n",
                               message="Added zoo")
    # Repo structure:
    #
    # @  3 Added zoo
    # |
    # o  2 Rename bar to zar
    # |
    # o  1 Changes bar
    # |
    # o  0 Add bar
    #

    # case 1: when a file renamed
    resp = commit_delta(left_cid=ctx1.hex(), right_cid=ctx2.hex())
    assert len(resp) == 1
    resp = resp[0]
    assert resp.from_path == b'bar'
    assert resp.to_path == b'zar'
    assert resp.old_mode == resp.new_mode

    # case 2: when a new file added
    resp = commit_delta(left_cid=ctx2.hex(), right_cid=ctx3.hex())
    assert len(resp) == 1
    resp = resp[0]
    assert resp.from_path == resp.to_path
    assert resp.old_mode == 0
    assert resp.new_mode == 0o100644

    # case 3: test with paths
    resp = commit_delta(left_cid=ctx0.hex(), right_cid=ctx3.hex(),
                        paths=[b'zoo'])
    assert len(resp) == 1
    resp = resp[0]
    assert resp.from_path == resp.to_path
    assert resp.from_path == b'zoo'

    # case 3.a
    resp = commit_delta(left_cid=ctx1.hex(), right_cid=ctx2.hex(),
                        paths=[b'zar'])
    assert len(resp) == 1
    resp = resp[0]
    assert resp.from_path == b'zar'
    assert resp.to_path == b'zar'
    assert resp.old_mode == 0
    assert resp.new_mode == 0o100644

    # case 3.b
    resp = commit_delta(left_cid=ctx1.hex(), right_cid=ctx2.hex(),
                        paths=[b'bar'])
    assert len(resp) == 1
    resp = resp[0]
    assert resp.from_path == b'bar'
    assert resp.to_path == b'bar'
    assert resp.old_mode == 0o100644
    assert resp.new_mode == 0

    # case 4: when commit_id does not correspond to a commit
    sha_not_exists = b'deadnode' * 5
    with pytest.raises(grpc.RpcError) as exc_info:
        commit_delta(sha_not_exists, ctx3.hex())
    assert exc_info.value.code() == grpc.StatusCode.FAILED_PRECONDITION


def test_diff_stats(diff_fixture):
    diff_stats = diff_fixture.diff_stats
    wrapper = diff_fixture.repo_wrapper

    wrapper.commit_file('bar', content="I am in\nrab\n",
                        message="Add bar")
    ctx1 = wrapper.commit_file('bar', content="I am in\nbar\n",
                               message="Changes bar")
    wrapper.command(b'mv', wrapper.repo.root + b'/bar',
                    wrapper.repo.root + b'/zar')
    ctx2 = wrapper.commit([b'bar', b'zar'], message="Rename bar to zar")
    ctx3 = wrapper.commit_file('zoo', content="I am in\nzoo\n",
                               message="Added zoo")
    # Repo structure:
    #
    # @  3 Added zoo
    # |
    # o  2 Rename bar to zar
    # |
    # o  1 Changes bar
    # |
    # o  0 Add bar
    #

    # case 1: when a file renamed
    resp = diff_stats(left_cid=ctx1.hex(), right_cid=ctx2.hex())
    assert len(resp) == 1
    resp = resp[0]
    assert resp.old_path == b'bar'
    assert resp.path == b'zar'
    assert not resp.additions
    assert not resp.deletions

    # case 2: when a new file added
    resp = diff_stats(left_cid=ctx2.hex(), right_cid=ctx3.hex())
    assert len(resp) == 1
    resp = resp[0]
    assert not resp.old_path
    assert resp.path == b'zoo'
    assert resp.additions == 2
    assert not resp.deletions

    # case 3: when commit_id does not correspond to a commit
    sha_not_exists = b'deadnode' * 5
    with pytest.raises(grpc.RpcError) as exc_info:
        diff_stats(sha_not_exists, ctx3.hex())
    assert exc_info.value.code() == grpc.StatusCode.FAILED_PRECONDITION


def test_find_changed_paths(diff_fixture):
    wrapper = diff_fixture.repo_wrapper

    (wrapper.path / 'sub').mkdir()
    (wrapper.path / 'sub/foo').write_text('foo content')
    (wrapper.path / 'bar').write_text('bar content')
    (wrapper.path / 'zoo').write_text('zoo content')
    ctx0 = wrapper.commit(rel_paths=['sub/foo', 'bar', 'zoo'],
                          add_remove=True)
    (wrapper.path / 'too').write_text('too content')
    (wrapper.path / 'too').chmod(0o755)
    (wrapper.path / 'sub/foo').write_text('foo content modified')
    (wrapper.path / 'bar').unlink()
    wrapper.command(b'cp', wrapper.repo.root + b'/zoo',
                    wrapper.repo.root + b'/zaz')
    ctx1 = wrapper.commit(rel_paths=['sub/foo', 'bar', 'zaz', 'too'],
                          add_remove=True)

    # Actual test (for ctx1)
    resp_dict = {
        b'too': [(ChangedPaths.Status.ADDED,
                  (OBJECT_MODE_DOES_NOT_EXIST, OBJECT_MODE_EXECUTABLE),
                  )],
        b'sub/foo': [(ChangedPaths.Status.MODIFIED, None)],
        b'bar': [(ChangedPaths.Status.DELETED,
                  (OBJECT_MODE_NON_EXECUTABLE, OBJECT_MODE_DOES_NOT_EXIST),
                  )],
        b'zaz': [(ChangedPaths.Status.COPIED, None)],
    }
    for changed in (diff_fixture.find_changed_paths,  # deprecated
                    diff_fixture.find_changed_paths_commits,  # new style
                    ):
        assert changed(commits=[ctx1.hex()]) == resp_dict
        assert changed(commits=[ctx1.hex()]) == resp_dict

        # when commit_id does not correspond to a commit
        wrong_cid = b'12face12' * 5
        with pytest.raises(grpc.RpcError) as exc_info:
            changed(commits=[wrong_cid])
        assert exc_info.value.code() == grpc.StatusCode.NOT_FOUND
        assert sysstr(wrong_cid) in exc_info.value.details()

    # Testing with two commits
    # in particular reproducing a former loop masking bug
    ctx2 = wrapper.commit_file('sub/foo', content='foo remodified')
    resp_dict[b'sub/foo'].append((ChangedPaths.Status.MODIFIED, None))
    assert diff_fixture.find_changed_paths_commits(
        [ctx1.hex(), ctx2.hex()]) == resp_dict

    # Testing by passing "parents" (changesets to compare to)
    ctx3 = wrapper.commit_file("toto")
    assert diff_fixture.find_changed_paths_commits(
        [ctx3.hex()], compare_to=[ctx1.hex()]
        ) == {
            b'sub/foo': [(ChangedPaths.Status.MODIFIED, None)],
            b'toto': [(ChangedPaths.Status.ADDED,
                       (OBJECT_MODE_DOES_NOT_EXIST,
                        OBJECT_MODE_NON_EXECUTABLE),
                       )],
        }

    # Tree requests
    repo = wrapper.repo
    sub0_oid, sub1_oid = [tree_oid(repo, ctx.hex().decode('ascii'), b'sub')
                          for ctx in (ctx0, ctx1)]
    assert diff_fixture.find_changed_paths_tree(sub0_oid, sub1_oid) == {
        b'foo': [(ChangedPaths.Status.MODIFIED, None)]
    }

    # TreeRequests on different paths. Arguably a lack in HGitaly, since
    # git diff-tree will happily issue provide a diff between arbitrary tree
    # objects. At least we're testing that our error raising works
    with pytest.raises(grpc.RpcError) as exc_info:
        # with our purely conventional oid, it does not matter if the
        # path exists or not for this test
        bogus_oid = tree_oid(repo, ctx0.hex().decode('ascii'), b'other')
        diff_fixture.find_changed_paths_tree(sub0_oid, bogus_oid)

    assert exc_info.value.code() == grpc.StatusCode.UNIMPLEMENTED


def test_find_changed_paths_copy_in_tree(diff_fixture):
    wrapper = diff_fixture.repo_wrapper

    (wrapper.path / 'subdir').mkdir()  # avoid all lengths to be 3
    (wrapper.path / 'subdir/bar').write_text('some bar')
    ctx0 = wrapper.commit(rel_paths=['subdir'], add_remove=True)

    wrapper.command(b'cp', wrapper.repo.root + b'/subdir/bar',
                    wrapper.repo.root + b'/subdir/baz')
    ctx1 = wrapper.commit(rel_paths=['subdir'])

    repo = wrapper.repo

    sub0_oid, sub1_oid = [tree_oid(repo, ctx.hex().decode('ascii'), b'subdir')
                          for ctx in (ctx0, ctx1)]
    assert diff_fixture.find_changed_paths_tree(sub0_oid, sub1_oid) == {
        b'baz': [(ChangedPaths.Status.COPIED, None)],
    }


def test_get_patch_id(diff_fixture):
    wrapper = diff_fixture.repo_wrapper
    patch_id = diff_fixture.get_patch_id

    hex0 = wrapper.commit_file('regular', content='foo\n').hex()
    hex1 = wrapper.commit_file('regular', content='bar\n').hex()
    hex2 = wrapper.commit_file(
        'foo.gz',
        content=b'\x1f\x8b\x08\x08\xd6Q\xc1e\x00\x03foo\x00K'
        b'\xcb\xcf\xe7\x02\x00\xa8e2~\x04\x00\x00\x00',
        message='gzipped file is easy binary').hex()
    hex3 = wrapper.commit_file(
        'foo.gz',
        content=b"\x1f\x8b\x08\x088=\xc2e\x00"
        b"\x03foo\x00K\xcb7\xe2\x02\x00\xb1F'q\x04\x00\x00\x00",
        message='gzipped file is easy binary').hex()

    patch_id1 = patch_id(hex0, hex1)
    patch_id2 = patch_id(hex1, hex2)
    patch_id3 = patch_id(hex2, hex3)

    # symbolic revisions work
    assert patch_id(hex2, b'branch/default') == patch_id3

    assert len(set((patch_id1, patch_id3, patch_id2))) == 3

    #
    # error cases
    #

    # Git not found
    wrapper.write_hgrc(dict(hgitaly={'git-executable': '/does/not/exist'}))
    with pytest.raises(grpc.RpcError) as exc_info:
        patch_id(hex0, hex1)
    exc = exc_info.value
    assert exc.code() == StatusCode.INTERNAL
    assert 'not found' in exc.details()

    # Git found, but not executable
    wrapper.write_hgrc(dict(
        hgitaly={'git-executable': str(wrapper.path / 'regular')}))
    with pytest.raises(grpc.RpcError) as exc_info:
        patch_id(hex0, hex1)
    exc = exc_info.value
    assert exc.code() == StatusCode.INTERNAL
    assert 'not executable' in exc.details()

    # Unknown revisions
    with pytest.raises(grpc.RpcError) as exc_info:
        patch_id(b'unknown', hex1)
    assert exc_info.value.code() == StatusCode.INTERNAL

    with pytest.raises(grpc.RpcError) as exc_info:
        patch_id(hex0, b'unknown')
    assert exc_info.value.code() == StatusCode.INTERNAL



# Copyright 2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later

import pytest
from mercurial_testhelpers import (
    as_bytes,
)
from heptapod.testhelpers import (
    LocalRepoWrapper,
)
from .common import (
    MINIMAL_HG_CONFIG,
)
from ..tag import (
    iter_gitlab_tags,
    iter_gitlab_tags_as_refs,
)


@pytest.fixture
def repo_wrapper(tmpdir):
    wrapper = LocalRepoWrapper.init(tmpdir, config=MINIMAL_HG_CONFIG)
    yield wrapper.repo, wrapper


def make_tag(wrapper, tag, rev=b'.'):
    wrapper.command('tag', as_bytes(tag), rev=as_bytes(rev))


def test_iteration(repo_wrapper):
    repo, wrapper = repo_wrapper
    ctx = wrapper.commit_file('foo')
    sha = ctx.hex().decode()
    assert list(iter_gitlab_tags(repo)) == []
    tag_name = b'important'
    tag_ref = b'refs/tags/important'

    make_tag(wrapper, 'important')
    assert list(iter_gitlab_tags(repo)) == [(tag_name, ctx)]
    assert list(iter_gitlab_tags(repo, deref=False)) == [(tag_name, sha)]
    assert list(iter_gitlab_tags_as_refs(repo)) == [(tag_ref, ctx)]
    assert list(iter_gitlab_tags_as_refs(repo, deref=False)) == [
        (tag_ref, sha)]



# Copyright 2021-2022 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import re
import subprocess

from mercurial import (
    commands,
    pycompat,
    ui as uimod,
)

from .stub.mercurial_repository_pb2 import (
    HeptapodConfigSection,
)

AutoPublish = HeptapodConfigSection.AutoPublish

GITLAB_PROJECT_FULL_PATH_FILENAME = b'gitlab.project_full_path'
MAIN_HGRC_FILE = b'hgrc'
MANAGED_HGRC_FILE = b'hgrc.managed'
INCLUDE_MANAGED_HGRC_LINE = b'%include ' + MANAGED_HGRC_FILE
HGRC_INCL_INHERIT_RX = re.compile(br'^%include (.*)/hgrc$', re.MULTILINE)
HEPTAPOD_CONFIG_SECTION = b'heptapod'
# TODO would be nice to import these from hgext3rd.heptapod
HEPTAPOD_CONFIG_AUTO_PUBLISH = b'auto-publish'
HEPTAPOD_CONFIG_ALLOW_BOOKMARKS = b'allow-bookmarks'
HEPTAPOD_CONFIG_ALLOW_MULTIPLE_HEADS = b'allow-multiple-heads'
AUTO_PUBLISH_MAPPING = {
    b'without-topic': AutoPublish.WITHOUT_TOPIC,
    b'nothing': AutoPublish.NOTHING,
    b'all': AutoPublish.ALL,
}
AUTO_PUBLISH_REVERSE_MAPPING = {v: k for k, v in AUTO_PUBLISH_MAPPING.items()}


def unbundle(repo, bundle_path: str, rails_sync=False):
    """Call unbundle with proper options and conversions.

    :param bool rails_sync: if ``True``, let the synchronization with the
       Rails app proceed (conversion to Git if needed, pre/post-receive hooks,
       etc.)
    """
    # make sure obsmarker creation is allowed while unbundle
    overrides = {(b'experimental', b'evolution'): b'all',
                 }
    if not rails_sync:
        overrides[(b'hooks', b'pretxnclose.heptapod_sync')] = b''

    # TODO it would be nice to have UPSTREAM a method
    # to unbundle from an arbitrary file-like object rather than
    # paths forcing us to dump to disk
    with repo.ui.configoverride(overrides, b'hgitaly.unbundle'):
        commands.unbundle(repo.ui, repo, pycompat.sysbytes(bundle_path))


def config_inherits(repo) -> bool:
    """Tell whether repo inherits configuration from its group."""
    hgrc = repo.vfs.tryread(MAIN_HGRC_FILE)
    if not hgrc:  # can be None or empty string for missing files
        return False

    return HGRC_INCL_INHERIT_RX.search(hgrc) is not None


def config_message_bool_from_ui(fields, ui, protocol_key, config_key,
                                as_recorded=False):
    default_opts = {}
    if as_recorded:
        default_opts['default'] = None

    value = ui.configbool(HEPTAPOD_CONFIG_SECTION, config_key, **default_opts)
    if value is None:
        return

    fields[protocol_key] = value


def heptapod_ui_config(ui, as_recorded=False) -> HeptapodConfigSection:
    """Convert the config from `ui` instance to message.

    :param as_recorded: if ``True``, values that are not explicitly set in
      config are not returned.
    """
    fields = {}

    auto_pub = ui.config(HEPTAPOD_CONFIG_SECTION, HEPTAPOD_CONFIG_AUTO_PUBLISH)

    if auto_pub is not None or not as_recorded:
        fields['auto_publish'] = AUTO_PUBLISH_MAPPING.get(auto_pub)

    for field, key in (
            ('allow_multiple_heads', HEPTAPOD_CONFIG_ALLOW_MULTIPLE_HEADS),
            ('allow_bookmarks', HEPTAPOD_CONFIG_ALLOW_BOOKMARKS),
    ):
        config_message_bool_from_ui(fields, ui, field, key,
                                    as_recorded=as_recorded)

    return HeptapodConfigSection(**fields)


def heptapod_config(repo) -> HeptapodConfigSection:
    """Return the `heptapod` configuration section, as seen by Mercurial.

    These are the fully resolved value, hence taking Group inheritance into
    account.
    """
    return heptapod_ui_config(repo.ui)


def heptapod_local_config(repo) -> HeptapodConfigSection:
    """Return the configuration spectific to this repository.

    It ignores global settings and those from the main HGRC, hence in
    particular what could be inherited from a Group.
    """
    ui = uimod.ui()
    ui.readconfig(repo.vfs.join(MANAGED_HGRC_FILE), trust=True)
    return heptapod_ui_config(ui, as_recorded=True)


def replace_heptapod_managed_config(repo, items, by_line):
    """Entirely replace the managed HGRC file with the given items.

    :param items: a dict, whose keys and values are as fields of the
      :class:`HeptapodConfigSection` message
    """
    with repo.wlock():
        with repo.vfs(MANAGED_HGRC_FILE,
                      mode=b'wb',
                      atomictemp=True,
                      checkambig=True) as fobj:
            fobj.write(b"# This file is entirely managed by Heptapod \n")
            fobj.write(b"# latest update ")
            fobj.write(by_line.encode('utf-8'))
            fobj.write(b"\n\n")

            if not items:
                return

            fobj.write(b"[heptapod]\n")
            auto_pub = items.get('auto_publish')
            if auto_pub is not None:
                fobj.write(b"auto-publish = ")
                fobj.write(AUTO_PUBLISH_REVERSE_MAPPING[auto_pub])
                fobj.write(b"\n")

            for field in ('allow_bookmarks', 'allow_multiple_heads'):
                val = items.get(field)
                if val is None:
                    continue

                hg_key = field.replace('_', '-')
                fobj.write(f'{hg_key} = {val}\n'.encode('ascii'))


def ensure_managed_config_inclusion(repo):
    with repo.wlock():
        for line in repo.vfs.tryread(MAIN_HGRC_FILE).splitlines():
            if line.strip() == INCLUDE_MANAGED_HGRC_LINE:
                return

        with repo.vfs(MAIN_HGRC_FILE,
                      mode=b'ab',
                      atomictemp=True,
                      checkambig=True) as fobj:
            fobj.write(INCLUDE_MANAGED_HGRC_LINE)
            fobj.write(b"\n")


def set_managed_config(repo, heptapod: HeptapodConfigSection,
                       remove_items, by_line):
    existing = heptapod_local_config(repo)
    items = {}
    for field in ('allow_bookmarks', 'allow_multiple_heads', 'auto_publish'):
        if field in remove_items:
            continue

        if heptapod.HasField(field):
            items[field] = getattr(heptapod, field)
        elif existing.HasField(field):
            items[field] = getattr(existing, field)
    replace_heptapod_managed_config(repo, items, by_line)
    ensure_managed_config_inclusion(repo)


def set_config_inheritance(repo, hgrc_dir, by_line):
    """Set inheritance from an HGRC file in given hgrc_dir.

    :param hgrc_dir: if None, the inheritance is removed if present.
      Otherwise has to be a relative path from the repo's wdir and
      if inheritance is absent, it is set to the file named ``hgrc``
      in the given ``hgrc_dir``.

    Tis currrenl
    """
    by_line = by_line.encode('utf-8')
    changed = False
    with repo.wlock():
        lines = repo.vfs.tryread(MAIN_HGRC_FILE).splitlines()
        for i, line in enumerate(lines):
            if HGRC_INCL_INHERIT_RX.search(line) is not None:
                if hgrc_dir is None:
                    lines[i] = b'# inheritance removed ' + by_line
                    changed = True
                break
        else:
            if hgrc_dir is not None:
                lines.insert(0,
                             b'%%include %s/hgrc' % hgrc_dir.encode('utf-8'))
                lines.insert(0, b'# inheritance restored ' + by_line)
                changed = True

        if not changed:
            return

        with repo.vfs(MAIN_HGRC_FILE,
                      mode=b'wb',
                      atomictemp=True,
                      checkambig=True) as fobj:
            for line in lines:
                fobj.write(line)
                fobj.write(b'\n')


def repo_size(repo):
    """Return repository size in kilobytes"""
    out = subprocess.check_output(('du', '-ks', os.fsdecode(repo.path)))
    return int(out.split()[0].decode('ascii'))



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import os
from pathlib import Path
import sys


def hg_exe_path():  # pragma no cover
    """Return a Mercurial executable path consistent with these tests.

    The executable should be consistent with the `mercurial` package imported
    in the tests.

    Most of these tests don't need to invoke Mercurial subprocesses, but
    some will have to (SSH push). In that case, we need a path to a version
    that is equivalent enough. A simple heuristic to achieve that is to try and
    return the path to the Mercurial executable that was installed with the
    `mercurial` package (e.g. from the same virtualenv etc.).

    If that is not enough, the `HGITALY_TESTS_HG_EXE_PATH` environment
    variable can be set to something suitable.
    """
    from_env = os.environ.get('HGITALY_TESTS_HG_EXE_PATH')
    if from_env:
        return from_env

    return str(Path(sys.executable).parent / 'hg')



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from mercurial import (
    pycompat,
)
from heptapod.testhelpers import (
    LocalRepoWrapper,
)

from ..stub.shared_pb2 import (
    CommitAuthor,
    Repository,
    SignatureType,
)
from ..stub.repository_pb2 import FetchBundleRequest

from .. import message


def test_tag(tmpdir):
    wrapper = LocalRepoWrapper.init(tmpdir)
    repo = wrapper.repo

    ctx = wrapper.write_commit('foo', message="The tagged chgs")

    # the factory function doesn't even need the tag to actually exist
    tag = message.tag(b'v3.2.1', ctx)
    assert tag.name == b'v3.2.1'
    assert not tag.id
    assert not tag.message
    assert not tag.message_size
    assert tag.tagger == CommitAuthor()
    assert tag.signature_type == SignatureType.NONE

    # we'll need a real tagging changeset
    wrapper.command('tag', b'v3.2.1', rev=ctx.hex(),
                    message=b'Setting the tag',
                    )
    tagging_ctx = repo[b'.']

    tag = message.tag(b'v3.2.1', ctx,
                      tagging=tagging_ctx,
                      signature_type=SignatureType.PGP)

    assert tag.name == b'v3.2.1'
    assert pycompat.sysbytes(tag.id) == tagging_ctx.hex()
    assert tag.message == b'Setting the tag'
    assert tag.message_size == 15
    assert tag.tagger == message.commit_author(tagging_ctx)
    assert tag.signature_type == SignatureType.PGP


def test_commit_author():

    class FakeContext:
        def __init__(self, user, timestamp_tz):
            self._user = user
            self._date = timestamp_tz

        def date(self):
            return self._date

        def user(self):
            return self._user

    timestamp = 1626786076.0  # as good as any, took it from a changeset

    name_and_email = b"My Self <me@name.test>"
    author = message.commit_author(FakeContext(name_and_email,
                                               (timestamp, 0)))
    assert author.name == b'My Self'
    assert author.email == b'me@name.test'
    assert author.timezone == b'+0000'
    assert author.date.seconds == int(timestamp)

    def with_tz(offset):
        return message.commit_author(FakeContext(name_and_email,
                                                 (timestamp, offset)))

    assert with_tz(-7200).timezone == b'+0200'
    assert with_tz(-16200).timezone == b'+0430'
    assert with_tz(3600).timezone == b'-0100'
    assert with_tz(4500).timezone == b'-0115'


def test_logging_class():
    log_msg = message.Logging(Repository(storage_name='thestore',
                                         relative_path='a/b/c'))
    assert repr(log_msg) == (
        'Repository { storage_name: "thestore" relative_path: "a/b/c" }'
    )


def test_logging_class_long_bytes():
    log_msg = message.Logging(FetchBundleRequest(data=b'\x01' * 100))
    assert repr(log_msg) == (
        r'FetchBundleRequest { data: "\001\001\001\001\001\001\001\001\001\001'
        r'\001\001\001\001\001\001\001\001\001\001\001\001\001\001\001\001\001'
        r'\001\001\001\001\001\001\001\001\001\001\001\001\001\001\001\001\001'
        r'\001\001\001\001\001\001... (truncated, actual size 100 bytes)" }'
    )



from hgitaly.stub.repository_pb2 import RepositoryExistsRequest
from hgitaly.stub.shared_pb2 import Repository
from hgitaly.stub.repository_pb2_grpc import RepositoryServiceStub


def wait_server_accepts_connection(gitaly_channel, timeout=5):
    """Wait for a server using the Gitaly protocol to accept connections.

    wait for server to start enough that the address in bound
    Then the `wait_for_ready` option is there to ensure full readiness, but
    the client must at least receive something, e.g., the `CONNECTING` status
    for that to work.
    """
    repo_stub = RepositoryServiceStub(gitaly_channel)
    repo_stub.RepositoryExists(
        RepositoryExistsRequest(
            repository=Repository(
                relative_path="we/dont/care/waiting/for/any/connection",
                storage_name="default")
        ),
        timeout=timeout,
        wait_for_ready=True
    )



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import mercurial_operations_pb2 as mercurial__operations__pb2


class MercurialOperationsServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.MergeAnalysis = channel.unary_unary(
                '/hgitaly.MercurialOperationsService/MergeAnalysis',
                request_serializer=mercurial__operations__pb2.MergeAnalysisRequest.SerializeToString,
                response_deserializer=mercurial__operations__pb2.MergeAnalysisResponse.FromString,
                )
        self.PublishChangeset = channel.unary_unary(
                '/hgitaly.MercurialOperationsService/PublishChangeset',
                request_serializer=mercurial__operations__pb2.PublishChangesetRequest.SerializeToString,
                response_deserializer=mercurial__operations__pb2.PublishChangesetResponse.FromString,
                )
        self.Censor = channel.unary_unary(
                '/hgitaly.MercurialOperationsService/Censor',
                request_serializer=mercurial__operations__pb2.CensorRequest.SerializeToString,
                response_deserializer=mercurial__operations__pb2.CensorResponse.FromString,
                )
        self.MergeBranch = channel.unary_unary(
                '/hgitaly.MercurialOperationsService/MergeBranch',
                request_serializer=mercurial__operations__pb2.MergeBranchRequest.SerializeToString,
                response_deserializer=mercurial__operations__pb2.MergeBranchResponse.FromString,
                )


class MercurialOperationsServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def MergeAnalysis(self, request, context):
        """Provide all information for a possible merge

        Will be used in mergeability checks and last-minute checks if needed.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PublishChangeset(self, request, context):
        """Publish a changeset

        As any publication in Mercurial, it also publishes all ancestors that
        were not already public.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Censor(self, request, context):
        """Censor a changeset

        Equivalent of `hg censor`
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def MergeBranch(self, request, context):
        """MergeBranch plays the same role as OperationService.UsermergeBranch with the following differences:

        - It is not a 2-phase operation, hence not streaming
        - The Request message has a slighly different set of options
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_MercurialOperationsServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'MergeAnalysis': grpc.unary_unary_rpc_method_handler(
                    servicer.MergeAnalysis,
                    request_deserializer=mercurial__operations__pb2.MergeAnalysisRequest.FromString,
                    response_serializer=mercurial__operations__pb2.MergeAnalysisResponse.SerializeToString,
            ),
            'PublishChangeset': grpc.unary_unary_rpc_method_handler(
                    servicer.PublishChangeset,
                    request_deserializer=mercurial__operations__pb2.PublishChangesetRequest.FromString,
                    response_serializer=mercurial__operations__pb2.PublishChangesetResponse.SerializeToString,
            ),
            'Censor': grpc.unary_unary_rpc_method_handler(
                    servicer.Censor,
                    request_deserializer=mercurial__operations__pb2.CensorRequest.FromString,
                    response_serializer=mercurial__operations__pb2.CensorResponse.SerializeToString,
            ),
            'MergeBranch': grpc.unary_unary_rpc_method_handler(
                    servicer.MergeBranch,
                    request_deserializer=mercurial__operations__pb2.MergeBranchRequest.FromString,
                    response_serializer=mercurial__operations__pb2.MergeBranchResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'hgitaly.MercurialOperationsService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class MercurialOperationsService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def MergeAnalysis(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/hgitaly.MercurialOperationsService/MergeAnalysis',
            mercurial__operations__pb2.MergeAnalysisRequest.SerializeToString,
            mercurial__operations__pb2.MergeAnalysisResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PublishChangeset(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/hgitaly.MercurialOperationsService/PublishChangeset',
            mercurial__operations__pb2.PublishChangesetRequest.SerializeToString,
            mercurial__operations__pb2.PublishChangesetResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Censor(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/hgitaly.MercurialOperationsService/Censor',
            mercurial__operations__pb2.CensorRequest.SerializeToString,
            mercurial__operations__pb2.CensorResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def MergeBranch(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/hgitaly.MercurialOperationsService/MergeBranch',
            mercurial__operations__pb2.MergeBranchRequest.SerializeToString,
            mercurial__operations__pb2.MergeBranchResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: analysis.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import lint_pb2 as lint__pb2
from . import shared_pb2 as shared__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0e\x61nalysis.proto\x12\x06gitaly\x1a\nlint.proto\x1a\x0cshared.proto\"\xaa\x01\n\x1a\x43heckBlobsGeneratedRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x36\n\x05\x62lobs\x18\x02 \x03(\x0b\x32\'.gitaly.CheckBlobsGeneratedRequest.Blob\x1a&\n\x04\x42lob\x12\x10\n\x08revision\x18\x01 \x01(\x0c\x12\x0c\n\x04path\x18\x02 \x01(\x0c\"\x83\x01\n\x1b\x43heckBlobsGeneratedResponse\x12\x37\n\x05\x62lobs\x18\x01 \x03(\x0b\x32(.gitaly.CheckBlobsGeneratedResponse.Blob\x1a+\n\x04\x42lob\x12\x10\n\x08revision\x18\x01 \x01(\x0c\x12\x11\n\tgenerated\x18\x02 \x01(\x08\x32}\n\x0f\x41nalysisService\x12j\n\x13\x43heckBlobsGenerated\x12\".gitaly.CheckBlobsGeneratedRequest\x1a#.gitaly.CheckBlobsGeneratedResponse\"\x06\xfa\x97(\x02\x08\x02(\x01\x30\x01\x42\x34Z2gitlab.com/gitlab-org/gitaly/v16/proto/go/gitalypbb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'analysis_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z2gitlab.com/gitlab-org/gitaly/v16/proto/go/gitalypb'
  _CHECKBLOBSGENERATEDREQUEST.fields_by_name['repository']._options = None
  _CHECKBLOBSGENERATEDREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _ANALYSISSERVICE.methods_by_name['CheckBlobsGenerated']._options = None
  _ANALYSISSERVICE.methods_by_name['CheckBlobsGenerated']._serialized_options = b'\372\227(\002\010\002'
  _globals['_CHECKBLOBSGENERATEDREQUEST']._serialized_start=53
  _globals['_CHECKBLOBSGENERATEDREQUEST']._serialized_end=223
  _globals['_CHECKBLOBSGENERATEDREQUEST_BLOB']._serialized_start=185
  _globals['_CHECKBLOBSGENERATEDREQUEST_BLOB']._serialized_end=223
  _globals['_CHECKBLOBSGENERATEDRESPONSE']._serialized_start=226
  _globals['_CHECKBLOBSGENERATEDRESPONSE']._serialized_end=357
  _globals['_CHECKBLOBSGENERATEDRESPONSE_BLOB']._serialized_start=314
  _globals['_CHECKBLOBSGENERATEDRESPONSE_BLOB']._serialized_end=357
  _globals['_ANALYSISSERVICE']._serialized_start=359
  _globals['_ANALYSISSERVICE']._serialized_end=484
# @@protoc_insertion_point(module_scope)



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later

from io import BytesIO
import os
import subprocess

import pytest

from heptapod.testhelpers import (
    LocalRepoWrapper,
)

from .. import diff as hgitaly_diff
from ..diff import (
    git_patch_id,
    run_git_patch_id,
    write_diff_to_file,
)
from ..tests.common import (
    MINIMAL_HG_CONFIG,
)
from hgitaly.service.diff import (
    CurrDiff,
    Limits,
    Parser,
)


def test_parser_corner_cases():
    parser = Parser(Limits(), CurrDiff())
    parser.parse([b""])


def diff_lines(cs_from, cs_to, **kw):
    out = BytesIO()
    write_diff_to_file(out, cs_from, cs_to, **kw)
    return out.getvalue().splitlines()


def test_git_patch_id(tmpdir):
    wrapper = LocalRepoWrapper.init(tmpdir, config=MINIMAL_HG_CONFIG)

    wrapper.commit_file('regular', content='foo\n')
    regular = (tmpdir / 'regular')
    regular.write('bar\n')

    script = tmpdir / 'script'
    script.write('#!/usr/bin/env python2\n'
                 'print "Hello, world"\n')
    script.chmod(0o755)
    cs1 = wrapper.commit([], add_remove=True)
    cs2 = wrapper.commit_file(
        'foo.gz',
        content=b'\x1f\x8b\x08\x08\xd6Q\xc1e\x00\x03foo\x00K'
        b'\xcb\xcf\xe7\x02\x00\xa8e2~\x04\x00\x00\x00',
        message='gzipped file is easy binary')

    assert diff_lines(cs1, cs2, dump_binary=False) == [
        b'diff --git a/foo.gz b/foo.gz',
        b'new file mode 100644',
        (b'index 0000000000000000000000000000000000000000..'
         b'%s_Zm9vLmd6' % cs2.hex()),
        b'Binary file foo.gz has changed',
        b'--- a/foo.gz',
        b'+++ b/foo.gz',
        b'@@ -1 +1 @@',
        b'-0000000000000000000000000000000000000000',
        b'+' + cs2[b'foo.gz'].hex(),
    ]

    patch_id2 = git_patch_id('git', cs1, cs2)

    cs3 = wrapper.commit_file(
        'foo.gz',
        content=b"\x1f\x8b\x08\x088=\xc2e\x00"
        b"\x03foo\x00K\xcb7\xe2\x02\x00\xb1F'q\x04\x00\x00\x00",
        message='gzipped file is easy binary')
    patch_id3 = git_patch_id('git', cs2, cs3)
    assert patch_id3 != patch_id2

    cs4 = wrapper.commit_file(
        'foo.gz',
        content=b"\x1f\x8b\x08\x08\xf2=\xc2e\x00"
        b"\x03foo\x00K\xcb7\xe6\x02\x00\xf0w<h\x04\x00\x00\x00",
        message='gzipped file is easy binary')
    patch_id4 = git_patch_id('git', cs3, cs4)
    assert len(set((patch_id4, patch_id3, patch_id2))) == 3


def test_run_git_patch_id_errors(tmpdir, monkeypatch):
    fake_git = tmpdir / 'git'
    fake_git_path = os.fsencode(str(fake_git))
    fake_git.write_text('\n'.join(('#!/usr/bin/env python3',
                                   'import sys',
                                   'import time',
                                   'timeout = sys.stdin.read()',
                                   'if not timeout:'
                                   '  sys.exit(21)',
                                   'time.sleep(float(timeout))',
                                   )),
                        encoding='ascii')
    fake_git.chmod(0o755)
    with pytest.raises(RuntimeError) as exc_info:
        run_git_patch_id(fake_git_path, lambda stdin: None)
    assert 'code 2' in exc_info.value.args[0]

    monkeypatch.setattr(hgitaly_diff, 'GIT_PATCH_ID_TIMEOUT_SECONDS', 0.001)

    with pytest.raises(subprocess.TimeoutExpired) as exc_info:
        run_git_patch_id(fake_git_path, lambda stdin: stdin.write(b'0.01'))



# Copyright 2024 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import shutil

from grpc import (
    RpcError,
    StatusCode
)
import pytest

from hgitaly.oid import (
    blob_oid,
)
from hgitaly.stub.analysis_pb2 import (
    CheckBlobsGeneratedRequest,
)
from hgitaly.stub.analysis_pb2_grpc import AnalysisServiceStub

from .fixture import ServiceFixture


class AnalysisFixture(ServiceFixture):

    stub_cls = AnalysisServiceStub

    def check_blobs_generated(self, rev_paths, **kw):
        Blob = CheckBlobsGeneratedRequest.Blob
        kw.setdefault('repository', self.grpc_repo)
        return [
            (blob.revision.decode(), blob.generated)
            for resp in self.rpc(
                    'CheckBlobsGenerated',
                    iter([CheckBlobsGeneratedRequest(
                        blobs=(Blob(revision=rp[0].encode(), path=rp[1])
                               for rp in rev_paths),
                        **kw)]))
            for blob in resp.blobs
        ]


@pytest.fixture
def analysis_fixture(grpc_channel, server_repos_root):
    with AnalysisFixture(grpc_channel, server_repos_root) as fixture:
        yield fixture


def test_check_blobs_generated(analysis_fixture):
    fixture = analysis_fixture
    wrapper = fixture.repo_wrapper

    sub = (wrapper.path / 'sub')
    sub.mkdir()

    cargo_path = b'sub/Cargo.lock'
    (wrapper.path / cargo_path.decode()).write_text("Content will not matter")
    xcode_path = b'something.nib'
    (wrapper.path / xcode_path.decode()).write_text("Content will not matter")
    (wrapper.path / 'foo').write_text("foo")

    stub_path = b'commit_pb2.py'
    import hgitaly.stub.commit_pb2 as stub_exmp
    shutil.copy(stub_exmp.__file__, wrapper.path / stub_path.decode())

    changeset = wrapper.commit(rel_paths=[cargo_path, stub_path, 'foo'],
                               add_remove=True,
                               message="Bunch of files")
    hg_sha = changeset.hex().decode()

    cargo_oid = blob_oid(wrapper.repo, hg_sha, cargo_path)
    xcode_oid = blob_oid(wrapper.repo, hg_sha, xcode_path)
    foo_oid = blob_oid(wrapper.repo, hg_sha, b'foo')
    stub_oid = blob_oid(wrapper.repo, hg_sha, stub_path)

    res = fixture.check_blobs_generated(((cargo_oid, cargo_path),
                                         (xcode_oid, xcode_path),
                                         (foo_oid, b'foo'),
                                         (stub_oid, stub_path)))
    assert res == [(cargo_oid, True), (xcode_oid, True),
                   (foo_oid, False), (stub_oid, True)]

    # call with the csid:path syntax works as well
    def blob_rev_path(csid, path):
        return (':'.join((csid, path.decode())), path)

    res = fixture.check_blobs_generated((blob_rev_path(hg_sha, cargo_path),
                                         blob_rev_path(hg_sha, b'foo'),
                                         blob_rev_path(hg_sha, stub_path)))
    assert [r[1] for r in res] == [True, False, True]

    # formally valid not resolving oids
    missing_path = b'unknown'
    missing_oid = blob_oid(wrapper.repo, hg_sha, missing_path)
    with pytest.raises(RpcError) as exc_info:
        fixture.check_blobs_generated([(missing_oid, missing_path)])
    assert exc_info.value.code() == StatusCode.INTERNAL

    arbitrary_sha1 = '12fe34ca' * 5
    missing_oid = blob_oid(wrapper.repo, arbitrary_sha1, missing_path)
    with pytest.raises(RpcError) as exc_info:
        fixture.check_blobs_generated([(missing_oid, missing_path)])
    assert exc_info.value.code() == StatusCode.INTERNAL

    # invalid oid
    missing_oid = '123'
    with pytest.raises(RpcError) as exc_info:
        fixture.check_blobs_generated([(missing_oid, missing_path)])
    assert exc_info.value.code() == StatusCode.INTERNAL



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import attr
from contextlib import contextmanager
import grpc
import os
import subprocess

from hgitaly.testing.grpc import wait_server_accepts_connection

from . import (
    GITALY_BIN_DIR,
    GITALY_INSTALL_DIR,
)


@attr.s
class GitalyServer:
    home_dir = attr.ib()

    def configure(self):
        home_dir = self.home_dir
        self.gitaly_conf = home_dir / 'gitaly_config.toml'
        self.gitaly_socket = home_dir / 'gitaly.socket'

        # this is required even if we won't use it (we're not sending any
        # GitLab hooks)
        gitlab_shell_dir = home_dir / 'gitlab-shell'
        gitlab_shell_dir.mkdir()

        default_storage = home_dir / 'default'
        default_storage.mkdir()
        conf_lines = [
            'socket_path = "%s"' % self.gitaly_socket,
            # Gitaly compilation outputs its binaries at the root
            # of the checkout
            'bin_dir = "%s"' % GITALY_BIN_DIR,
            '[gitlab-shell]',
            'dir = "%s"' % gitlab_shell_dir,
            '[gitaly-ruby]',
            'dir = "%s"' % (GITALY_INSTALL_DIR / 'ruby'),
            '[[storage]]',
            'name = "default"',
            'path = "%s"' % default_storage,
            '[git]',
            'use_bundled_binaries = true',
            ''
        ]
        self.gitaly_conf.write_text("\n".join(conf_lines))

    @contextmanager
    def start(self):
        self.configure()

        env = dict(os.environ)
        env['GITALY_TESTING_NO_GIT_HOOKS'] = "1"
        timeout = int(env.pop('GITALY_STARTUP_TIMEOUT', '30').strip())

        # as of Gitaly 14.9, the `logging.dir` setting does not have
        # it write to a file, let's redirect ourselves.
        with open(self.home_dir / 'gitaly.log', 'w') as logf:
            gitaly = subprocess.Popen(
                [GITALY_BIN_DIR / 'gitaly', self.gitaly_conf],
                stdout=logf, stderr=logf,
                env=env)

        with grpc.insecure_channel('unix:' + str(self.gitaly_socket),
                                   ) as gitaly_channel:
            wait_server_accepts_connection(gitaly_channel, timeout=timeout)
            yield gitaly_channel

        gitaly.terminate()
        gitaly.wait()



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Fixtures for Gitaly comparison tests.

Some people don't like the ``conftest.py`` magic, and it is true it is
possible to share fixtures amongst test modules by direct imports instead.

But direct imports of fixtures have a severe drawback: if fixture `b` depends
on fixture `a`, then importing `b` is not enough. Implicitely, pytest will
look for `a` where `b` is imported, so that actually both have to be imported,
even if the tests using `b` don't have `a` in their signatures.

This can probably cost developers a few hours of debugging, so not fighting
the framework seems the right thing to do, here.

Another drawback (minor) is that pyflakes doesn't understand that an imported
fixture is actually used leading to these errors::

  from .gitaly import gitaly_channel   # F401 (not used)

  def test_something(gitaly_channel):  # F811 (redefinition of unused import)
      pass

To keep the magic low, we have most of the logic in the `gitaly` and
`comparison` modules.
"""
import pytest
from .gitaly import GitalyServer
from .rhgitaly import RHGitalyServer
from .comparison import gitaly_comparison_fixture
from .hgitaly_rhgitaly_comparison import hgitaly_rhgitaly_comparison_fixture


@pytest.fixture(scope='module')
def gitaly_channel(server_repos_root):
    with GitalyServer(server_repos_root).start() as channel:
        yield channel


@pytest.fixture(scope='module')
def rhgitaly_channel(server_repos_root, grpc_addr):
    # grpc_addr is of the `localhost:PORT` form
    sidecar_address = 'http://' + grpc_addr
    with RHGitalyServer(server_repos_root,
                        sidecar_address=sidecar_address,
                        ).running() as channel:
        yield channel


@pytest.fixture()
def gitaly_comparison(server_repos_root,  # module scope
                      gitaly_channel,  # module scope
                      grpc_channel,  # module scope
                      monkeypatch,  # function scope
                      ):
    with gitaly_comparison_fixture(server_repos_root,
                                   gitaly_channel,
                                   grpc_channel,
                                   monkeypatch,
                                   ) as comparison:
        yield comparison


@pytest.fixture()
def gitaly_rhgitaly_comparison(server_repos_root,  # module scope
                               gitaly_channel,  # module scope
                               rhgitaly_channel,  # module scope
                               grpc_channel,  # module scope
                               monkeypatch,  # function scope
                               ):
    with gitaly_comparison_fixture(server_repos_root,
                                   gitaly_channel,
                                   grpc_channel,
                                   monkeypatch,
                                   rhgitaly_channel=rhgitaly_channel,
                                   ) as comparison:
        yield comparison


@pytest.fixture()
def rhgitaly(server_repos_root,  # module scope
             rhgitaly_channel,  # module scope
             monkeypatch,  # function scope
             ):
    with hgitaly_rhgitaly_comparison_fixture(
            server_repos_root=server_repos_root,
            rhgitaly_channel=rhgitaly_channel,
            monkeypatch=monkeypatch,
    ) as fixture:
        yield fixture


@pytest.fixture()
def hgitaly_rhgitaly_comparison(server_repos_root,  # module scope
                                rhgitaly_channel,  # module scope
                                grpc_channel,  # module scope
                                monkeypatch,  # function scope
                                ):  # pragma no cover, remove when used again
    with hgitaly_rhgitaly_comparison_fixture(
            server_repos_root,
            hgitaly_channel=grpc_channel,
            rhgitaly_channel=rhgitaly_channel,
            monkeypatch=monkeypatch,
    ) as comparison:
        yield comparison


@pytest.fixture()
def hgitaly_rhgitaly_comparison_no_repo(server_repos_root,  # module scope
                                        rhgitaly_channel,  # module scope
                                        grpc_channel,  # module scope
                                        monkeypatch,  # function scope
                                        ):
    with hgitaly_rhgitaly_comparison_fixture(
            server_repos_root,
            hgitaly_channel=grpc_channel,
            rhgitaly_channel=rhgitaly_channel,
            skip_repo=True,
    ) as comparison:
        yield comparison



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from enum import Enum
from io import BytesIO
import re

from mercurial import (
    error,
    node as nodemod,
    pycompat,
)
from hgext3rd.evolve import headchecking
from heptapod.gitlab.branch import (
    branchmap_branch_from_gitlab_branch,
    gitlab_branch_from_branchmap_branch,
    parse_wild_gitlab_branch,
    gitlab_branch_ref,
)
from hgext3rd.heptapod.branch import (
    GITLAB_BRANCHES_MISSING,
    get_default_gitlab_branch,
    gitlab_branches,
)
from hgext3rd.heptapod import (
    ensure_gitlab_branches as hg_git_ensure_gitlab_branches,
)
from .gitlab_ref import (
    path_match_any,
    sort_key_ref_changeset_date_asc,
    sort_key_ref_changeset_date_desc,
)


import logging
logger = logging.getLogger(__name__)


class BranchSortBy(Enum):
    """Sort criteria for branches.

    As long as Gitaly does not define a common set of sort criteria
    in `shared.proto`, the enums they use might differ with each request.
    Hence we need to remap to an inner class such as this one.

    Also, we might in the future introduce specific variations, depending on
    conventions for branch representation (which currently give the
    lexicographical ordering putting regular named branches before topics but
    are prone to change)
    """
    FULL_REF_NAME = 0
    UPDATED_ASC = 1
    UPDATED_DESC = 2


def _extract_branchmap_heads(repo, entry,
                             avoid_bookmark_shadowing=False):
    """From the given branchmap entry, extract default head and list of heads.

    The default head is what needs to be returned for a simple branchmap
    branch lookup. It can be ``None`` if all heads are invisible to GitLab
    (closed or bookmarked).

    :param avoid_bookmark_shadowing: if ``True`` the bookmarked revision with
       the highest revno can be returned if there's nothing else
    :return: (default head, all heads visible to GitLab), all given as
             :class:`changectx` instances.
    """
    # branchmap entries results are already sorted by increasing rev number
    revs = [repo[node].rev() for node in entry]
    contexts = (repo[rev]
                for rev in headchecking._filter_obsolete_heads(repo, revs))
    visible = [c for c in contexts if not (c.closesbranch() or c.bookmarks())]

    if not visible and avoid_bookmark_shadowing:
        # rare case, performance is irrelevant, we can re-lookup
        for ctx in (repo[node] for node in reversed(entry)):
            if not ctx.obsolete() and not ctx.closesbranch():
                return ctx, [ctx]

    if not visible:
        return None, visible

    return visible[-1], visible


def changeset_branchmap_entry(repo, ctx):
    """Return the branchmap entry for branch/topic combination of a changeset

    :param ctx: the changeset, given as a :class:`changectx`.
    """
    branch, topic = ctx.branch(), ctx.topic()
    key = branch if not topic else b':'.join((branch, topic))
    return repo.branchmap()[key]


def gitlab_branch_head(repo, gl_branch):
    """Return the unique head of the given GitLab branch.

    Does not resolve other types of "revisions".

    :return: a :class:`changectx` or ``None`` if there's no changeset for
             the given ``gl_branch``.
    """
    if not len(repo):
        return None

    gl_branches = gitlab_branches(repo)
    if gl_branches is not GITLAB_BRANCHES_MISSING:
        sha = gl_branches.get(gl_branch)
        if sha is None:
            return None
        try:
            return repo[sha]
        except error.RepoLookupError as exc:
            logger.warning("GitLab branches state file for repo %r "
                           "has a bogus changeset %r for %r: %r", repo.path,
                           sha, gl_branch, exc)
            return None

    logger.warning("gitlab_branch_head for %r: no GitLab branches state file "
                   "defaulting to slow direct analysis.", repo.root)
    wild_hex = parse_wild_gitlab_branch(gl_branch)
    if wild_hex is not None:
        wild_bin = nodemod.bin(wild_hex)
        try:
            wild_ctx = repo[wild_bin]
        except error.RepoLookupError:
            return None

        if wild_ctx.bookmarks():
            # a bookmarked head is never wild
            return None

        heads = _extract_branchmap_heads(
            repo, changeset_branchmap_entry(repo, wild_ctx))[1]
        return wild_ctx if len(heads) > 1 and wild_ctx in heads else None

    branchmap_key = branchmap_branch_from_gitlab_branch(gl_branch)
    if branchmap_key is not None:
        try:
            entry = repo.branchmap()[branchmap_key]
        except KeyError:
            return None
        return _extract_branchmap_heads(
            repo, entry,
            avoid_bookmark_shadowing=(
                gl_branch == get_default_gitlab_branch(repo)),
        )[0]

    # last chance: bookmarks
    bookmark_node = repo._bookmarks.get(gl_branch)
    if bookmark_node is None:
        return None
    return repo[bookmark_node]


def iter_gitlab_branches(repo, deref=True):
    """Iterate on all visible GitLab branches

    Each iteration yields a pair ``(gl_branch, target)`` where ``gl_branch``
    is the name of the GitLab branch and ``target`` is a :class:`changectx`
    instance or a hexadecimal node id, as a :class:`str` instance.

    :param bool deref: if ``True``, resolves targets as :class:`changectx`
      instance, ignoring lookup errors. If ``False``, yield targets as
      hexadecimal node ids.
    """
    if not len(repo):
        return

    gl_branches = gitlab_branches(repo)
    if gl_branches is not GITLAB_BRANCHES_MISSING:
        for branch, sha in gl_branches.items():
            if deref:
                try:
                    yield branch, repo[sha]
                except error.RepoLookupError as exc:
                    logger.error("Unknown changeset ID in GitLab branches "
                                 "statefile for repo at %r: %r",
                                 repo.root, exc)
            else:
                yield branch, sha.decode('ascii')
    else:
        logger.warning("iter_gitlab_branches for %r: "
                       "no GitLab branches state file "
                       "defaulting to slow direct analysis.", repo.root)
        for branch, target in iter_compute_gitlab_branches(repo):
            if not deref:
                target = target.hex().decode('ascii')
            yield branch, target


def gitlab_branches_matcher(patterns):
    """Returns a compiled regexp for pattern matching on GitLab branches

    This is meant to match the reference Go implementation in Gitaly's
    `UpdateRemoteMirror`, which in turn is fed the protected branches pattern.
    Even though we won't implement the same gRPC method (too Git centric), it
    is taken as reference for how glob patterns on branch names are interpreted
    in Gitaly. As of Gitaly 14.4.1::

      // newReferenceMatcher returns a regexp which matches references that should
      // be updated in the mirror repository. Tags are always matched successfully.
      // branchMatchers optionally contain patterns that are used to match branches.
      // The patterns should only include the branch name without the `refs/heads/`
      // prefix. "*" can be used as a wilcard in the patterns. If no branchMatchers
      // are specified, all branches are matched successfully.
      func newReferenceMatcher(branchMatchers [][]byte) (*regexp.Regexp, error) {
          sb := &strings.Builder{}
          sb.WriteString("^refs/tags/.+$|^refs/heads/(")

          for i, expression := range branchMatchers {
              segments := strings.Split(string(expression), "*")
              for i := range segments {
                  segments[i] = regexp.QuoteMeta(segments[i])
              }

              sb.WriteString(strings.Join(segments, ".*"))

              if i < len(branchMatchers)-1 {
                  sb.WriteString("|")
              }
          }

          if len(branchMatchers) == 0 {
              sb.WriteString(".+")
          }

          sb.WriteString(")$")

          return regexp.Compile(sb.String())
      }

    Difference in our implementation:

    - the returned pattern will be made to match branch
      names directly instead of refs. We don't have to take care of tags.
    - we return ``None`` if ``patterns`` is empty
    """  # noqa: E501  (long quote from Golang code)
    pattern_seen = False

    re_pattern = BytesIO()
    re_pattern.write(b'^')
    for pattern in patterns:
        if pattern_seen:
            re_pattern.write(b'|')
        pattern_seen = True
        re_pattern.write(b'(')
        re_pattern.write(b'.*'.join(
            re.escape(segment) for segment in pattern.split(b'*')))
        re_pattern.write(b')')

    if pattern_seen:
        re_pattern.write(b'$')
    return re.compile(re_pattern.getvalue())


def iter_gitlab_branches_matching(repo, patterns):
    rx = gitlab_branches_matcher(patterns)
    # OPTIM this has the drawback of resolving the changeset contexts
    # for branches that don't match. Not expected to be a performance
    # problem at this stage, but can be further optimized by performing
    # the pattern matching in `iter_gitlab_branches()`
    return ((name, ctx) for name, ctx in iter_gitlab_branches(repo)
            if rx.match(name) is not None)


def iter_compute_gitlab_branches(repo):
    """Generator that computes GitLab branches from scratch"""
    for key, entry in pycompat.iteritems(repo.branchmap()):
        gl_branch = gitlab_branch_from_branchmap_branch(key)
        default, visible = _extract_branchmap_heads(
            repo, entry,
            avoid_bookmark_shadowing=(
                gl_branch == get_default_gitlab_branch(repo)))
        if not visible:
            continue

        yield gl_branch, default
        if len(visible) > 1:
            for head in visible:
                yield b'wild/' + head.hex(), head

    for key, node in pycompat.iteritems(repo._bookmarks):
        yield key, repo[node]


def iter_gitlab_branches_as_refs(repo, patterns=None, **kwargs):
    """Same as :func:`iter_gitlab_branches`, yielding full Git ref paths.
    """
    for branch, target in iter_gitlab_branches(repo, **kwargs):
        ref_path = gitlab_branch_ref(branch)
        if patterns and not path_match_any(ref_path, patterns):
            continue

        yield gitlab_branch_ref(branch), target


def sorted_gitlab_branches_as_refs(repo, after=None, sort_by=None, deref=True):
    """Return all visible, GitLab branches, sorted.

    :param bytes after: full ref name of a branch. If not `None`,
       return only branches which occur strictly after the given one in the
       sort ordering used.
    :param bool deref: same as in :func:`iter_gitlab_branches`, except
       that dereferencing may happen internally if the sort criterion
       requires it.
    :params kwargs: passed over to :func:`iter_gitlab_branches`
    :returns: iterable of pairs (full Git ref name, changectx or str).
      Can be just an iterator.

    By default GitLab sorts branches lexicographically.
    Since Heptapod is using `/` separated special names,
    it would make sense to perform a variation over that
    (named branches first, then topics, lexicographically sorted
    for each, of course).
    """
    # TODO OPTIM (would be premature until we have some profile figures
    # pointing at this function)
    # Depending on the total size of ref_chgsets, it might be better to
    # - do any linear filtering before sorting
    # - resolve changecontexts only while yielding, a bigger refactoring
    #   also involving `iter_gitlab_branches()`.
    if sort_by == BranchSortBy.FULL_REF_NAME:
        sort_key = None
    elif sort_by == BranchSortBy.UPDATED_ASC:
        sort_key = sort_key_ref_changeset_date_asc
    elif sort_by == BranchSortBy.UPDATED_DESC:
        sort_key = sort_key_ref_changeset_date_desc

    inner_deref = deref if sort_by == BranchSortBy.FULL_REF_NAME else True

    ref_targets = sorted(iter_gitlab_branches_as_refs(repo, deref=inner_deref),
                         key=sort_key)
    if after is not None:
        ref_targets = iter(ref_targets)
        for ref, _ in ref_targets:
            if ref == after:
                break

    if inner_deref and not deref:
        ref_targets = ((ref, target.hex().decode('ascii'))
                       for ref, target in ref_targets)
    yield from ref_targets


def ensure_gitlab_branches_state_file(repo):
    """Make sure that the GitLab branches state file exists.

    Some implementations have to rely on it, and we will soon basically
    require in in practice, lest the performance degradation become
    intolerable.

    For now, we can assume that all repositories are actually handled with
    the mirroring to Git code from hgext3rd.heptapod.

    When that becomes false, we'll be faced with two options:

    - have enough trust in our migration scenarios that we simply make it
      a hard requirement or
    - keep the code able to recompute all GitLab branches around, such as
      :func:`gitlab_branch_head` as of this writing and use it to reconstruct
      the file.

    See also py-heptapod#8
    """
    hg_git_ensure_gitlab_branches(repo.ui, repo)



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later

# flake8: noqa E402
import os
import time
from mercurial.i18n import _
from mercurial import (
    demandimport,
    error,
    pycompat,
    registrar,
    util,
)

demandimport.disable()

from hgitaly.server import (
    run_forever,
    UnsupportedUrlScheme,
    InvalidUrl,
    BindError,
)
from . import revset

if util.safehasattr(registrar, 'configitem'):

    configtable = {}
    configitem = registrar.configitem(configtable)
    configitem(b'hgitaly', b'tokei-executable', b'tokei')
    configitem(b'hgitaly', b'license-detector-executable', b'license-detector')
    configitem(b'hgitaly', b'git-executable', b'git')


cmdtable = {}
command = registrar.command(cmdtable)


revsetpredicate = revset.revsetpredicate


DEFAULT_LISTEN_URL = b'tcp://127.0.0.1:9237'


def reraise_abort(msg, exc):
    """Raise error.abort from inner exception.

    :param msg: a bytes string containing as many `%s` as needed.
    """
    fmt_args = tuple(pycompat.sysbytes(str(a)) for a in exc.args)
    raise error.Abort(_(b"hgitaly-serve: " + msg) % fmt_args)


@command(b'hgitaly-serve',
         [(b'', b'repositories-root',
           b'',
           _(b'Path to the root directory for repositories storage. '
             b'Defaults to the `heptapod.repositories-root` '
             b'configuration item '),
           _(b'REPOS_ROOT')),
          (b'', b'listen',
           [],
           _(b'URL to listen on, default is %s, can be '
             b'repeated to listen on several addresses' % DEFAULT_LISTEN_URL),
           _(b'ADDRESS')),
          (b'', b'mono-process',
           False,
           _(b'Mono process mode (useful for debugging, notably with pdb).'
             b'No forking occurs, the main process is the single worker. '
             b'In particular the hgitaly.workers configuration item is '
             b'ignored.')),
          ],
         _(b'[OPTIONS]...'),
         norepo=True
         )
def serve(ui, **opts):
    """Start a HGitaly server with only one storage, named 'default'

    By default the root of repositories is read from the
    `heptapod.repositories-root` configuration item if present.
    """
    # local time does not make sense for a server-oriented process,
    # so let's switch to UTC. Mercurial itself is transparent about that
    # server-side commits will be recorded as been done UTC, which is fine
    # (it does not change the recorded time, which is always UTC, just some
    # ways of displaying it).
    os.environ['TZ'] = 'UTC'
    time.tzset()
    listen_urls = [pycompat.sysstr(u) for u in opts['listen']]
    if not listen_urls:
        # Any default in the option declaration would be added to
        # explicely passed values. This is not what we want, hence the late
        # defaulting done here.

        # Although gRPC defaults to IPv6, existing Gitaly setups
        # (GDK and monolithic Docker container) bind explicitely on IPv4
        listen_urls = [DEFAULT_LISTEN_URL.decode()]
    repos_root = opts.get('repositories_root')
    if not repos_root:
        repos_root = ui.config(b'heptapod', b'repositories-root')
        if repos_root is None:
            raise error.Abort(_(
                b"No value found in configuration for "
                b"'heptapod.repositories-root'. Please define it or run "
                b"the command with the --repositories-root option."
                ))
    nb_workers = ui.configint(b'hgitaly', b'workers')
    run_opts = dict(nb_workers=nb_workers,
                    restart_done_workers=True,
                    mono_process=opts.get('mono_process'))
    max_rss_mib = ui.configint(b'hgitaly', b'worker.max-rss-mib')
    if max_rss_mib is not None:
        run_opts['max_rss_mib'] = max_rss_mib
    monitoring_interval = ui.configint(b'hgitaly',
                                       b'worker.monitoring-interval-seconds')
    if monitoring_interval is not None:
        run_opts['monitoring_interval'] = monitoring_interval

    shutdown_timeout = ui.configint(
        b'hgitaly',
        b'worker.graceful-shutdown-timeout-seconds')
    if shutdown_timeout is not None:
        run_opts['graceful_shutdown_timeout_seconds'] = shutdown_timeout

    storages = dict(default=repos_root)
    try:
        run_forever(listen_urls, storages, **run_opts)
    except UnsupportedUrlScheme as exc:
        reraise_abort(b"unsupported URL scheme: '%s'", exc)
    except InvalidUrl as exc:
        reraise_abort(b"invalid URL: '%s' (%s)", exc)
    except BindError as exc:
        reraise_abort(b"could not listen (bind) to URL '%s'", exc)
